################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "logger" 
require "erb"
require "devops_core/util"
require "devops_core/crypto_util"
require "devops_library/providers/library_system_provider"
require "devops_library/providers/mock_ram_library_provider"
require "devops_library/providers/ram_session"
require "devops_library/providers/package"
 

################################################################################
# Class Definition                                                             #
################################################################################

module DevopsLibrary
  module Providers
    class RAMLibraryProvider < LibrarySystemProvider
      include DevopsCore::Util
      
      QUOTE="\"" 
      
      ################################################################################
      # Class Methods
      ################################################################################
    
      # Returns the right RAMLibraryProvider class.  For local builds use
      # a mock provider that pretends to interact with RAM.
      #  
      # Parameters
      # ==========
      # - use_mock - (Boolean) If true, return a mock provider
      # 
      def RAMLibraryProvider.newFor(use_mock, args, build=nil, log=nil, log_level=Logger::WARN)
        if use_mock
          return DevopsLibrary::Providers::MockRAMLibraryProvider.new(args, build, log, log_level)
        else
          return DevopsLibrary::Providers::RAMLibraryProvider.new(args, build, log, log_level)
        end
      end
      
      ################################################################################
      # Instance Methods
      ################################################################################
    
       # Initialize the RAMLibraryProvider
      #  
      # Parameters
      # ==========
      #   args : hash of the name-value pair
      #      Required
      #         :repository_url : RAM server url for eg. http://localhost:9080/ram.ws
      #         :user : User ID to authenticate the ram server
      #         :password : Password to authenticate the ram server
      
      
      def initialize(args, build=nil, log=nil, log_level=Logger::WARN)
        
        @root_server_uri = get_property_value(args, "repository_url")
        @user = get_property_value(args, "user")  
        @password = DevopsCore::CryptoUtil.decrypt_if_needed("password", get_property_value(args, "password"), args["KEY_FILE"])
        
        @build = build
        
        if @root_server_uri.nil? 
          raise ArgumentError, "A value for :repository_url is required."
        end
  
        if @user.nil? 
          raise ArgumentError, "A value for :user is required."
        end
  
        if @password.nil? 
          raise ArgumentError, "A value for :password is required."
        end
        
        if log.nil?
          log = Logger.new(STDOUT) 
        end
        @log = log
        @log_level=log_level
        @log.level=log_level      
        
        @ram_session = RAMSession.new(@root_server_uri, @user, @password, @log, @log_level)

      end

      # Creates a RAM asset and optionally publishes an artifact(s) to it.
      #  
      # Parameters
      # ==========
      #   package_id      : GUID of the asset
      #   package_version : Version of the asset
      #   metadata        : Additional metadata that can be used to create the asset
      #        build_url       : URL of the build
      #        name            : Application name
      #        description     : description of the asset to be created. 
      #                          If description is not passed, name would be set as description
      #        community_name  : Name of the community
      #        asset_type      : Asset Type of the asset
      #        attributes      : array of name / value
      #                          attribute_name : name of the attribute
      #                          attribute_value : value of the attribute
      #        classifications : classifications for the asset
      #
      #   packages : Hash containing the files to be published to the asset
      #              fully qualified file name => file name
      #   optimize : Flag signifying whether or not asset creation is optimized.  If the
      #              value is true then the asset is only created if the artifacts to be
      #              included in this new asset are different than the artifacts included
      #              in an existing asset (if one exists).  If the value is false then
      #              the asset is always created and the artifacts published to it.  The
      #              default value is true.
      #   
      # Output
      # ======
      #    Package : DevopsLibrary::Providers::Package 
      #
      def create_package(package_id, package_version, metadata={}, packages=nil, optimize=true)
        package_id = handle_spaces(package_id)

        # If there are no artifacts to publish or optimization is turned off then
        # create the asset and return
        if packages.nil? || packages.size == 0 || optimize == false
          return create_publish_package(package_id, package_version, packages, metadata)
        end
        
        # Calculate the checksum on the packages to be published to this asset
        aggregate_checksum = calculate_aggregate_checksum(packages)

        # Look for an existing asset with this guid and checksum.  Default
        # the query for builds other than personal builds.
        # Match GUID, state is not {rejected, personal, submitted}, and checksum matches
        query = "txt,id:(#{package_id}) !fState:(rejected) * !fState:(personal) * !fState:(submitted) * attribute:('Build Checksum'='#{aggregate_checksum}')"
        if !@build.type.nil?
          if @build.type == "Personal"
            query = "txt,id:(#{package_id}) !fState:(rejected) * attribute:('Build Checksum'='#{aggregate_checksum}')"
          end
        end
        query=URI.escape(query).gsub("(", "%28").gsub(")","%29")
        query="ramSearch:(1#{query})"
        
        matching_assets = query_packages(query, "ras:version", false)
       
        # query_packages may have returned several matching assets.  Since they
        # are sorted just use the first one.
        if matching_assets.length > 0
          # A matching asset was found.  Use it.
          package = matching_assets[0]
        else
          # No matching asset was found
          package = nil
        end
        
        if package.nil?
          # An asset with this GUID and checksum does not exists so create it
          package = create_publish_package(package_id, package_version, packages, metadata, aggregate_checksum)
        end

        return package
      end

      # Deletes a RAM asset.
      #  
      # Parameters
      # ==========
      #   package_id : GUID of the asset
      #   package_version : Version of the asset
      #   
      def delete_package(package_id, package_version)

        package_id = handle_spaces(package_id)

        response =  @ram_session.delete_asset(package_id, package_version)
        response
      end


      # Adds file to the asset
      # Parameters
      # ==========
      #   package : DevopsLibrary::Providers::Package  
      #   filename : filename of the published file
      #   file : file to be published
      # Output
      # ======
      #    url of the artifact
      #
      def publish_package(package, file, filename=nil) 
        return @ram_session.add_artifact(handle_spaces(package.id), package.version, file, filename)
      end

      # Downloads the asset to the folder and returns the path of the file
      # Parameters
      # ===========
      #   Required
      #       package_id : GUID of the asset
      #       package_version : Version of the asset
      #       folder : Directory to which the artifact should be downloaded
      #   Optional
      #       filename : Save As file name
      #
      def fetch_package(package_id, package_version, folder, filename=nil)
          return @ram_session.download_asset(handle_spaces(package_id), package_version, folder, filename)
      end 

      # Fetches the single artifact from the asset and downloads it to the folder
      # Parameters
      # ===========
      #   Required
      #       package_id : GUID of the asset
      #       package_version : Version of the asset
      #       filename : name of the artifact
      #       folder : Directory to which the artifact should be downloaded
      #
      #
      #   Return
      #       Absolute file path
      #
      def fetch_package_file(package_id, package_version, filename, folder)
          return @ram_session.download_artifact(handle_spaces(package_id), package_version, filename, folder)
      end 
      
      # Fetches the artifacts matching the regular expression from the asset and download it to the folder
      # Parameters
      # ===========
      #   Required
      #       package_id : GUID of the asset
      #       package_version : Version of the asset
      #       filename : name or regular expression matching the artifact
      #       folder : Directory to which the artifact should be downloaded
      #
      #
      #   Return
      #       Absolute file path
      #
      def fetch_package_files(package_id, package_version, filename, folder) 
          package=get_package(package_id, package_version)
          paths = Array.new 
          if !package.nil? and !package.data.nil? and !package.data['artifacts'].nil?
            filename = filename.gsub('*', '(.*)')
            package.data['artifacts'].each do |artifact|  
              if (artifact['title'] =~ /#{filename}/)
                  path = @ram_session.download_resource(handle_spaces(package_id), package_version, artifact['content']['resource'], folder)
                  paths.push(path)
              end 
            end
          end
            
          if !paths.any?
            raise "No artifacts found matching the name #{filename}"
          end
          
          return paths
      end 
      
      # Generates the deps file for the given asset
      # Parameters
      # ===========
      #   Required
      #       package : DevopsLibrary::Providers::Package
      #       
      #   Return
      #       url of the Deps file
      #
      def publish_dependency_file(package)
          response = @ram_session.get_asset(handle_spaces(package.id), package.version)
          asset_metadata = JSON.parse(response.body)
          artifact_urls = generate_dependency_content(asset_metadata)
          relationships = asset_metadata['relationships']
          if !relationships.nil? and relationships.length >0
              for i in 0...relationships.length 
                    relationship = relationships[i]
                    resource = relationship["resource"] 
                    relationship_type = relationship["type"]["value"]
                    if relationship_type == "dependency" 
                        #get asset details  
                        begin
                          response =  @ram_session.get_asset_from_url(resource) 
                          parsedData = JSON.parse(response.body) 
                          deps_content = generate_dependency_content(parsedData)
                          
                          if artifact_urls.nil?
                            artifact_urls=deps_content
                          else
                            artifact_urls="#{deps_content}\n#{artifact_urls}"
                          end 
                          
                       rescue Exception => e
                          #TODO: replace with logging
                          #puts e.message 
                       end 
                    end 
              end # end of for 
           end
           if !artifact_urls.nil?
             asset_name=asset_metadata["title"]
             asset_name=asset_name.gsub(' ', '_')
             @ram_session.add_artifact_content( handle_spaces(package.id), package.version, artifact_urls, "Deps-#{asset_name}")
             return get_url(package.id, package.version, "Deps-#{asset_name}")
           end
      end
      
      # Returns the related asset URL's as an array
      # Parameters
      # ===========
      #   Required
      #       package_id : GUID of the asset
      #       package_version : Version of the asset
      #       related_types: Relationship between the packages to download (default={'dependency', 'Infrastructure'})
      #       recursive : true or false (default= true)
      #          true : Recursive fetches the parents 
      #       
      #   Return
      #       Array of the related asset url's
      #
      def get_related(package_id, package_version, related_types=nil, recursive=true)
         package_id=handle_spaces(package_id)
         if related_types.nil?
           related_types=["dependency","Infrastructure"]
         end
         resource="oslc/assets/#{package_id}/#{package_version}"
         asset_resources = get_related_resources(resource, related_types, recursive)
         related_packages = nil
         if !asset_resources.nil?
           related_packages = Array.new
           for i in 0...asset_resources.length
             asset_resource = asset_resources[i]
             array = asset_resource.split('/')
             id=array[2]
             version=array[3]
             url = get_url(id,version)
             package = DevopsLibrary::Providers::Package.new(id, version, nil, url) 
             related_packages.push(package)
           end
         end
         return related_packages
      end
      
      # Resolves to the latest version of the package
      # Parameters
      # ===========
      #   Required
      #       guid     : GUID of the asset
      #       version  : Version of the asset, it could be an osgi format to specify version range
      #       metadata :
      #         metadata [:state] = state of the asset
      #         metadata [:name] = name of the asset
      #         metadata [:type] = type of the asset
      #         metadata [:query] = ram query          
      #         metadata [:category] = Array of categories.
      #   Return
      #       package : DevopsLibrary::Providers::Package
      #
      def resolve_package(package_id, package_version, metadata)
        state = nil
        type = nil
        name = nil
        query=nil
        category=nil
        community=nil
        unless metadata.nil?
          state = metadata[:state]
          if state.nil?
            state = metadata["state"]
          end
          name = metadata[:name]
          if name.nil?
            name = metadata["name"]
          end
          type = metadata[:type]
          if type.nil?
            type = metadata["type"]
          end
          query = metadata[:query]
          if query.nil?
            query = metadata["query"]
          end
          category = metadata[:category]
          if category.nil?
            category = metadata["category"]
          end
          community = metadata[:community]
          if community.nil?
            community = metadata["community"]
          end
        end
        
        if query.nil?
          query=create_query(package_id, package_version, state, type, name, category, community)
        else
          query="#{query}/sort=-ras:version"
        end
        
        #execute query 
        asset_properties="ras:id"
        response = @ram_session.execute_query(query, asset_properties)
        data = JSON.parse(response.body)
        results = data['results']
        if !results.nil? and results.length > 0
          result = results[0]
          idVal = result['id']  
          url = get_url(idVal['ras:GUID'], idVal['ras:version'])
          return DevopsLibrary::Providers::Package.new(idVal['ras:GUID'], idVal['ras:version'], result, url) 
        end
      end
      
      # Adds the attribute to the package
      # Parameters
      # ===========
      #   Required
      #       package : Package structure
      #             id      : GUID of the asset
      #             version : Version of the asset
      #       attribute_name : Name of the attribute
      #       attribute_value : Value of the attribute
      #       attribute_url : URL of the attribute
      #
      def add_attribute (package, attribute_name, attribute_value, attribute_url=nil) 
        if !attribute_url.nil?
            attribute_value =  "<a href='#{attribute_url}' target='_blank'>#{attribute_value}</a>" 
        end   
       
        attribute = @ram_session.get_attribute(attribute_name)
          
        attribute_binding=nil
        attribute_template_file = nil
        if !attribute.nil?
          #use attribute identifier  
          attribute_template_file = File.join(File.dirname(__FILE__), 'attributeTemplate.json') 
          attribute_identifier = attribute['identifier']
          attribute_href = attribute['href']
          attribute_binding = { 
              :NAME => attribute_name,
              :VALUES => attribute_value,
              :IDENTIFIER => attribute_identifier,
              :HREF => attribute_href
          }       
        else
          attribute_template_file = File.join(File.dirname(__FILE__), 'attributeTemplateSD.json')
          attribute_binding = { 
              :NAME => attribute_name,
              :VALUES => attribute_value
          }
        end
          
        attribute_json = evaluate(File.read(attribute_template_file), attribute_binding)
        @ram_session.add_attribute(handle_spaces(package.id), package.version, attribute_json)
      end
      
      # Publishes the artifact content link to the asset
      # Parameters
      # ==========
      #   package (structure)  
      #       id : ID of the package
      #       version : Version of the package
      #       metadata : additional metadata describing the package
      #       url : url of the package
      #
      #   url : url to be published as artifact
      #   name : [Optional] If name is nil, name will be the last segment of the url 
      # Output
      # ======
      #    url of the file in the package
      #
      def publish_link(package, url, name=nil)
         if name.nil?
            segments=url.split("/")
            name = segments.last
            if name.nil? || name==""
              name = segments[segments.length-2]
            end
        end  
         
        artifact_template_file = File.join(File.dirname(__FILE__), 'artifactTemplate.json')
        artifact_binding = { 
            :NAME => name,
            :URL => url
        }         
        artifact_json = evaluate(File.read(artifact_template_file), artifact_binding)
        @ram_session.add_artifact_link(handle_spaces(package.id), package.version, artifact_json)
        
        return get_url(package.id, package.version, name)
      end
      
      # Returns the package 
      # Parameters
      # ===========
      #   Required
      #       package_id : id of the package
      #       package_version : Version of the package
      #   Return
      #       package DevopsLibrary::Providers::Package
      #
      def get_package(package_id, package_version)
         package_id=handle_spaces(package_id)
         response = @ram_session.get_asset(package_id, package_version)
         data = JSON.parse(response.body)
         idVal = data['id'] 
         url = get_url(idVal['ras:GUID'], idVal['ras:version']) 
         return DevopsLibrary::Providers::Package.new(idVal['ras:GUID'], idVal['ras:version'], data, url)  
      end
       
      # Updates the package from the metadata
      # Parameters
      # ===========
      #   Required
      #       package_id : id of the package
      #       package_version : Version of the package
      #       metadata : asset metadata
      #
      def update_package(package_id, package_version, metadata)
         package_id=handle_spaces(package_id)
         @ram_session.update_asset(package_id, package_version, metadata)
      end
      
      # Get URL of the package / file
      #    package_id : ID of the package
      #    package_version : Version of the package
      #    file : artifact name
      #     
      def get_url(id, version, filename=nil)
        id=handle_spaces(id)
        if filename.nil?
          return @ram_session.get_url( id, version, filename)
        else
          package=get_package(id, version)
          if !package.nil? and !package.data.nil? and !package.data['artifacts'].nil?
             filename = filename.gsub('*', '(.*)')
             package.data['artifacts'].each do |artifact|  
               if (artifact['title'] =~ /#{filename}/)
                  if !artifact['reference']['kind'].nil? && artifact['reference']['kind']== "url"
                    return @ram_session.get_url(id, version, artifact['title'])
                  else
                    return @ram_session.get_resource_url(artifact['content']['resource'])
                  end
               end 
             end
          end
          raise "No artifacts found matching the name #{filename}"
        end
      end
      
      # Creates a relationship between two packages
      #    package_id : ID of the package
      #    package_version : Version of the package
      #    related_package_id : ID of the related package
      #    related_package_version : Version of the related package
      #    relationship : relationship name
      #   
      def relate_package(package_id, package_version, related_package_id, related_package_version, relationship)
       puts "In relate_package, package_id #{package_id}, package_version #{package_version}, relaged_package_id #{related_package_id}, related_package_version #{related_package_version}, relationship #{relationship}"
       package_id=handle_spaces(package_id)
       related_package_id=handle_spaces(related_package_id)
       relationship_template_file = File.join(File.dirname(__FILE__), 'relationshipTemplate.json')
          binding = { 
            :RELATED_GUID => related_package_id,
            :RELATED_VERSION => related_package_version,
            :RELATIONSHIP => relationship
       }         
       relationship_json = evaluate(File.read(relationship_template_file), binding)
       @ram_session.add_relationship(package_id, package_version, relationship_json)
     
      end
      
      # classifies the package with the classification
      #    package_id : ID of the package
      #    package_version : Version of the package
      #    category_path : category path
      #    recategorize : true, if sibling is exclusive and should be removed before setting the category
      #                 : false, if sibling is exclusive and attempt to categorize the asset will
      #                 : throw exception
      def categorize_package(package_id, package_version, category_path, recategorize=true)
        package_id=handle_spaces(package_id)
        schema=get_category_schema(category_path)
        if schema.nil?
          raise ArgumentError, "Category '#{category_path}' does not exist."
        end
        category = get_category(category_path,schema)
        if category.nil?
          raise ArgumentError, "Category '#{category}' does not exist."
        end
        category_id=category["identifier"]
        category_template_file = File.join(File.dirname(__FILE__), 'categoryTemplate.json')
        category_binding = { 
              :CATEGORY_ID => category_id
        }          
        category_json = evaluate(File.read(category_template_file), category_binding)
        
        parent_category= get_parent_category(category_path, schema)
        
        @ram_session.add_category(package_id, package_version, category_json, parent_category, recategorize)
      end
      
      # Queries server for the list of packages
      #    query : search query
      #    sort : sort criteria for the search
      #    ascending : true or false. Default is ascending
      #
      # Return
      #     packages as DevopsLibrary::Providers::Package
      def query_packages(query, sort=nil, ascending=true)
        if !sort.nil?
          if ascending
            query="#{query}/sort=#{sort}"
          else
            query="#{query}/sort=-#{sort}"
          end  
        end 
        response = @ram_session.execute_query(query)
        data = JSON.parse(response.body)
        results = data['results']
        asset_array = Array.new
        unless results.nil?
          results.each do |asset|
            idVal = asset['id']  
            url = get_url(idVal['ras:GUID'], idVal['ras:version'])
            package = DevopsLibrary::Providers::Package.new(idVal['ras:GUID'], idVal['ras:version'], asset, url) 
            asset_array.push(package)
          end
        end
        return asset_array
      end      

      # Sets the status of the package as deployed for the build environment. The environment is 
      # the environment set by the build
      # Parameters
      # ==========
      #   package (structure)  
      #       id : ID of the package
      #       version : Version of the package
      #       metadata : additional metadata describing the package
      #       url : url of the package
      #

      def mark_deployed(package)
         if @build.nil?
           raise ArgumentError, "Build not set during provider initialization"
         end
         if @build.environment.nil?
           raise ArgumentError, "Environment not set in the build"
         end
          
         category_path="Continuous Delivery/Deployment History/#{@build.environment}/Deployed" 
         categorize_package(package.id, package.version, category_path, false) 
      end
      
      # Sets the status of the package as test successful or test failed
      # Parameters
      # ==========
      #   package (structure)  
      #       id : ID of the package
      #       version : Version of the package
      #       metadata : additional metadata describing the package
      #       url : url of the package
      #
      #   test_status : DevopsLibrary::Providers::LibrarySystemProvider::TEST_SUCCESS or DevopsLibrary::Providers::LibrarySystemProvider::TEST_FAIL
      #
      #
      def mark_tested(package, test_status)
         if @build.nil?
           raise ArgumentError, "Build not set during provider initialization"
         end
         if @build.environment.nil?
           raise ArgumentError, "Environment not set in the build"
         end
          
         category_path="Continuous Delivery/Deployment History/#{@build.environment}/Tested/#{test_status}"
         begin
            categorize_package(package.id, package.version, category_path, false) 
         rescue Exception => ex
           if ex.message.include? "Cannot re-categorize the asset"
              @log.warn ex.message
           else
              raise ex
           end
         end
      end
      
      
      private
      ##############################################################################
      # Private Methods                                                            #
      ############################################################################## 
      
      def handle_spaces(id)
        unless id.nil?
          id=id.gsub(' ', '_')
        end
        return id
      end
      
      def create_package_from_json (asset_as_json)
         data = JSON.parse(asset_as_json)
         idVal = data['id'] 
         url = get_url(idVal['ras:GUID'], idVal['ras:version']) 
         return DevopsLibrary::Providers::Package.new(idVal['ras:GUID'], idVal['ras:version'], data, url) 
      end
      
      
      def generate_dependency_content(asset_metadata)
        artifacts = asset_metadata['artifacts']
        artifact_urls = nil
        #loop all the artifacts and create the url 
        if !artifacts.nil? and artifacts.length >0 
          for j in 0...artifacts.length 
             artifact = artifacts[j] 
             if !artifact['title'].start_with? "Deps-"
                resourceLink =  "#{@root_server_uri}/#{artifact['content']['resource']}"
                if artifact_urls.nil?
                  artifact_urls=resourceLink
                else
                  artifact_urls="#{resourceLink}\n#{artifact_urls}"
                end 
             end
          end 
        end
        return artifact_urls 
      end
      
      def get_related_resources(resource, related_types, recursive, asset_resources = nil)
          if asset_resources.nil?
            asset_resources = Array.new  
          end
           
          response = @ram_session.get_asset_from_url(resource)
          asset_metadata = JSON.parse(response.body)
          relationships = asset_metadata['relationships']
          if !relationships.nil? and relationships.length >0
              for i in 0...relationships.length 
                    relationship = relationships[i]
                    resource = relationship["resource"] 
                    relationship_type = relationship["type"]["value"]
                    if related_types.include?(relationship_type) and !asset_resources.include?(resource)
                        #get asset details  
                        begin
                          asset_resources.push(resource)
                          if recursive
                            get_related_resources(resource, related_types, recursive, asset_resources)  
                          end
                       rescue Exception => e
                          puts e.message 
                       end 
                    end 
              end # end of for 
           end
           return asset_resources
      end
      
      #resolve categories
      def resolve_categories(asset_metadata)
           ramCategories = asset_metadata["ramCategories"]
           unless ramCategories.nil?
              ramCategories.each { |category_entry|  
                category_path = category_entry["resource"]
                schema=get_category_schema(category_path)
                unless schema.nil?
                  category = get_category(category_path,schema)
                  unless category.nil?
                    category_entry["resource"]=category["identifier"]    
                  end
                end 
              }
           end
      end

      def resolve_relationship(asset_metadata) 
           #search asset based on name and version 
           relationships = asset_metadata["relationships"]
           unless relationships.nil? 
              for i in 0...relationships.length 
                  relationship = relationships[i]
                  resource = relationship["resource"]
                   
                  if (!resource.nil? and resource.kind_of?(Hash) )
                    id = resource["id"]
                    id=handle_spaces(id)
                    version = resource["version"]
                    state = resource ["state"]  
                    name = resource ["name"]
                    type = resource ["type"]
                    query=resource["query"]
                    category=resource["category"]
                    community=resource["community"]
                    if query.nil?
                      query=create_query(id,version,state, type, name, category,community)
                    else
                      query="#{query}/sort=-ras:version"
                    end
                    #execute query 
                    asset_properties="ras:id"
                    response = @ram_session.execute_query(query, asset_properties)
                    # 
                    parsed_data = JSON.parse(response.body)
                    if  !parsed_data.nil? 
                      results = parsed_data['results']
                      if !results.nil? and results.length > 0
                        first = results[0]
                        idVal = first['id']
                        guid =  idVal['ras:GUID']
                        version = idVal['ras:version']
                        relationship.delete("resource")
                        relationship["resource"] = "oslc/assets/#{guid}/#{version}" 
                      else
                        relationships[i] = nil
                      end
                    else
                      relationships[i] = nil
                    end 
                  else
                    relationship["resource"]=handle_spaces(resource)
                  end # end of if 
               end # end of for
               relationships.delete_if {|relationship| relationship == nil}
            end 
      end
      
      def create_query(id, version, state, type, name, categories,community)
          query = nil 
          id_query = nil
          version_query=nil 
          name_query=nil
          unless id.nil?
            id_query = "id:(#{id})"
          end
          
          unless version.nil?  
              version_query = nil
              if version.start_with? '[' or version.start_with? '(' or version.end_with? ')' or version.end_with? ']'
                  version=version.gsub(',', ' TO ').gsub('(', '{').gsub(')', '}')
                  version_query="version:(#{version})"
              elsif version != '*'  
                 version_query = "version:(#{version})"
              end 
          end # end of version
          
          if !id_query.nil? and !version_query.nil? 
            query="txt,#{id_query} #{version_query}" 
            query=URI.escape(query).gsub("(", "%28").gsub(")","%29")
            query="ramSearch:(1#{query})/sort=-ras:version"
            return query
          end
          
          unless name.nil? 
            name_query = "name:(#{name})"
          end
          
          if !id_query.nil? or !version_query.nil? or !name_query.nil?
            query="txt,#{id_query} #{version_query} #{name_query}" 
          end
          
          
          unless type.nil? 
             type=type.downcase.gsub(' ', '_') 
             if query.nil? 
                query = "fType,#{type}"
             else
                query = "#{query}$fType,#{type}"
            end 
          end #end of type
          
          if !state.nil?
             state=state.downcase.gsub(' ', '_') 
             if query.nil? 
                query = "fState,#{state}"
             else
                query = "#{query}$fState,#{state}"
            end 
          end #end of state
          
          if !community.nil?
             community=community.downcase.gsub(' ', '_') 
             if query.nil? 
                query = "fGroup,#{community}"
             else
                query = "#{query}$fGroup,#{community}"
            end 
          end #end of state
          
          unless categories.nil?
            category_entries=Array.new
            categories.each do |category_path|
                 # get the category schema
                 # if not added add that  
                 schema=get_category_schema(category_path)
                 schema_entry="Classification,#{schema['identifier']}#/"
                 if !category_entries.include?(schema_entry)
                   category_entries.push(schema_entry)
                 end
                 # get the sub category 
                 # add the category id
                 category=get_category(category_path,schema,true)
                 sub_category=get_category(category_path,schema)
                 sub_category_entry="#{category['name'].downcase.gsub(' ', '_')},#{sub_category['identifier'].split('#').last}"
                 if !category_entries.include?(sub_category_entry)
                   category_entries.push(sub_category_entry)
                 end
            end
            category_query=nil
            category_entries.each do |category_entry| 
              if category_query.nil?
                category_query=category_entry
              else
                category_query="#{category_query}$#{category_entry}"
              end
            end  
            if query.nil? 
                query = category_query
             else
                query = "#{query}$#{category_query}"
            end 
          end
          
          # set it to descending
          query=URI.escape(query).gsub("(", "%28").gsub(")","%29")
          query="ramSearch:(1#{query})/sort=-ras:version"
          return query
      end
      
      def add_attributes(asset_as_json, metadata)
         attributes = metadata[:attributes]
         unless attributes.nil?
            for i in 0...attributes.length 
                attribute = attributes[i]
                attribute_name = attribute[:name]
                attribute_value = attribute[:value]
                attribute_url = attribute[:url]
                
                unless attribute_value.nil? 
                   if !attribute_url.nil?
                     attribute_value =  "<a href='#{attribute_value}' target='_blank'>#{attribute_value}</a>" 
                   end  
                   asset_as_json = add_asset_attribute(asset_as_json, attribute_name, attribute_value) 
                end
             end
          end  
          return asset_as_json       
      end
      
  
     
      
      def create_hash_from_json(json)
      begin
        if json.is_a?(Hash)
          return json
        end      
              
        json_obj = nil
        json_obj = JSON.parse(json)
    
        return json_obj
    
      rescue Exception => e
    
        json_content = nil
    
        if File.file?(json) or !(json =~ URI::regexp).nil?
          json_content = open(json)
        else
          return nil
        end
    
        return JSON.parse(json_content)
      
      end
    end
    
    def add_build_attributes(asset_data)
       unless @build.nil?
         #Struct.new(:uri, :uuid, :id, :type)
         asset_data = merge_asset_attribute(asset_data, "Build Label", @build.id)
         asset_data = merge_asset_attribute(asset_data, "Build ID", @build.uuid)
         asset_data = merge_asset_attribute(asset_data, "Build Type", @build.type)
         asset_data = merge_asset_attribute(asset_data, "Build URL", "Build Project URL", @build.uri)
         asset_data = merge_asset_attribute(asset_data, "Build Requested By", @build.build_requestor_id) 
       end
       return asset_data
    end
    
    def merge_asset_attribute(asset_data, attribute_name, attribute_value, attribute_url=nil)
       unless @build.id.nil?
           attribute_data = get_attribute_data(attribute_name, attribute_value, attribute_url)
           if asset_data["ramAttributeValues"].nil?
             asset_data["ramAttributeValues"] = attribute_data["ramAttributeValues"]
           else
            asset_data["ramAttributeValues"] = asset_data["ramAttributeValues"] | attribute_data["ramAttributeValues"]
           end
       end
       return asset_data  
    end
    
    def add_build_category(asset_data)
       unless @build.nil? 
         if !@build.type.nil?
            category_path="Continuous Delivery/Build/#{@build.type}"
            asset_data = merge_build_category(asset_data, category_path)
         end
       end
       return asset_data
    end
    
    def merge_build_category(asset_data, category_path)
       unless category_path.nil?
           category_data = get_category_data(category_path)
           unless category_data.nil?
               if asset_data["ramCategories"].nil?
                 asset_data["ramCategories"] = category_data["ramCategories"]
               else
                asset_data["ramCategories"] = asset_data["ramCategories"] | category_data["ramCategories"]
               end
            end
       end
       return asset_data  
    end
    
    def get_category_data(category_path)
        schema=get_category_schema(category_path)
        unless schema.nil?
          category = get_category(category_path,schema)
          if !category.nil?
              category_id=category["identifier"]
              category_template_file = File.join(File.dirname(__FILE__), 'categoryTemplate.json')
              category_binding = { 
                  :CATEGORY_ID => category_id 
              }         
              category_json = evaluate(File.read(category_template_file), category_binding)
              return JSON.parse(category_json)   
          end
        end
       
    end
    
   
    
    def get_attribute_data(attribute_name, attribute_value, attribute_url=nil)
        if !attribute_url.nil?
            attribute_value =  "<a href='#{attribute_url}' target='_blank'>#{attribute_value}</a>" 
        end
     
        attribute = @ram_session.get_attribute(attribute_name)
        attribute_identifier = nil     
        if !attribute.nil?
          #use attribute identifier  
          attribute_identifier = attribute['identifier']
          attribute_href = attribute['href']
       end
            
       attribute_template_file = File.join(File.dirname(__FILE__), 'attributeTemplate.json')
       attribute_binding = { 
            :NAME => attribute_name,
            :VALUES => attribute_value,
            :IDENTIFIER => attribute_identifier,
            :HREF => attribute_href
       }         
       attribute_json = evaluate(File.read(attribute_template_file), attribute_binding)
       return JSON.parse(attribute_json)
    end
    
    def get_parent_category(category_path, schema)
      index = category_path.rindex('/') 
      parent_category_path=category_path[0,index]
      return get_category(parent_category_path, schema)
    end
    
    def get_category_schema(path)
       array = path.split('/')
       return @ram_session.get_category_schema(array[0])
    end
    
    def get_category(path, schema, only_category=false)
      categories=schema["category"]
      unless categories.nil?
          categories.each do |category|
            if path.start_with?(category["displayPath"])
                if path==category["displayPath"] or only_category
                  return category
                else
                  return get_sub_category(category["subCategory"], path)
                end
             end
           end # end of each loop
       end
       return nil
    end
    
    def get_sub_category(subCategories, path)
      unless subCategories.nil?
          subCategories.each do |subCategory|
            if path.start_with?(subCategory["displayPath"])
              if path==subCategory["displayPath"]
                return subCategory
              else
                return get_sub_category(subCategory["subCategory"], path)
              end
            end
           end # end of each loop
       end
       return nil
    end
    
    def add_asset_attribute(asset_as_json, attribute_name, attribute_value)
          asset_hash = JSON.parse(asset_as_json)
          
          attribute_template_file = File.join(File.dirname(__FILE__), 'attributeTemplate.json')
          attribute_binding = { 
            :NAME => attribute_name,
            :VALUES => attribute_value
          }         
          attribute_json = evaluate(File.read(attribute_template_file), attribute_binding)
          attribute_json_hash = JSON.parse(attribute_json)
          
          #attributes_hash is an array
          attributes_hash_array = asset_hash['ramAttributeValues']
          
          if attributes_hash_array.nil?
            attributes_hash_array = Array.new
            attributes_hash_array.push(attribute_json_hash)
            asset_hash = { # adds ramAttributeValues
              :ramAttributeValues=> attributes_hash_array
           }.merge(asset_hash) 
           
          else
            attributes_hash_array.push(attribute_json_hash)
            asset_hash['ramAttributeValues'] = attributes_hash_array
          end
 
          #hash to jSON
          return asset_hash.to_json
     end 
      
     def evaluate(template, context)
         # call utility evaluate template
         return DevopsCore::Util::evaluate_template(template, context)
    
      end
      
      def validate_metadata(metadata)
  
        if metadata['title'].nil?
          raise ArgumentError, "metadata['title'] must be set to a valid name"
        end
        
        
        if metadata['id']['ras:version'].nil?
          raise ArgumentError, "metadata['id']['ras:version'] must be set to a valid version identifier"
        end
        
        if metadata['type']['value'].nil?
          raise ArgumentError, "metadata['type']['value'] must be set to a valid asset type"
        end

        
        if metadata['community']['value'].nil?
          raise ArgumentError, "metadata['community']['value'] must be set to a valid commmunity name"
        end        
        
        
        if metadata['subject'].nil?
          raise ArgumentError, "metadata['subject'] must be set to a valid description"
        end 
                
      end 
      
      def get_property_value(args, name)
        value = args[name]
        if value.nil?
          value = args[name.to_sym]
        end
        return value
      end
      
      # Calculate the checksum for the contents of a zip file
      # Parameters
      # ==========
      #   file : The zip file
      # Output
      # ======
      #   The aggregate checksum for the contents of the zip file
      #
      def calculate_checksum_zip(file)
        filearray = Array.new
        
        # Open the zip/jar file and process each file in it
        Zip::ZipFile.open(file) { |zip_file|
          zip_file.each { |f|

            # If this file is a zip/jar then we need to process
            # the files in it as well 
            if f.name.end_with?(".jar") or f.name.end_with?(".zip") 
              crc = calculate_checksum_zip_entry(zip_file, f)
              filearray.push("#{crc}") 
            else
              filearray.push("#{f.crc}")  
            end
          }
        }
    
        # Sort the array so we can expect consistent results
        filearray = filearray.sort 
        
        # Create a checksum on the array.  That is what we will
        # use as the checksum for this zip/jar file.
        crc = Digest::MD5.hexdigest(filearray.inspect)
        
        return crc
      end

      # Calculate the checksum on a ZipEntry that happens to be
      # a zip/jar file.
      # Parameters
      # ==========
      #   zip_file : The zip file containing the ZipEntry
      #   entry    : The ZipEntry that is a zip/jar file
      # Output
      # ======
      #   The aggregate checksum for the entry
      #
      def calculate_checksum_zip_entry(zip_file, entry)

        # Create a temporary file to hold the contents of this ZipEntry
        temp_file= "#{Dir.tmpdir}/#{Time.now.to_f}.tmp"
        
        # Extract the ZipEntry to the temporary file
        zip_file.extract(entry, temp_file)
        
        # Compute the checksum on this zip/jar file
        crc = calculate_checksum_zip(temp_file)

        # Delete the temporary file
        File.delete(temp_file)
 
        return crc
      end

      # Calculate an aggregate checksum for the files in packages Hash.
      # Parameters
      # ==========
      #   packages : Hash containing the files to be published to the asset.
      #              fully qualified file name => file name
      # Output
      # ======
      #   The aggregate checksum for the files in the packages Hash.
      #
      def calculate_aggregate_checksum(packages)
        
        # Return a default value if no packages are to be published
        if packages.nil?
          return "0"
        end
        
        if packages.size == 0
          return "0"
        end

        filearray = Array.new
        
        # Calculate and save the checksum for each package
        packages.each do|file,filename|
          if file.end_with?(".jar") or file.end_with?(".zip") 
            checksum = calculate_checksum_zip(file)
          else
            checksum = Digest::MD5.hexdigest(File.read(file))
          end
          filearray.push(checksum)
        end
        # Sort the array of checksums so results are consistent
        filearray = filearray.sort

        # Get the checksum based on the sorted array of checksums.
        # This is the aggregate checksum.
        aggregate_checksum = Digest::MD5.hexdigest(filearray.inspect)

        return aggregate_checksum
      end
 
      # Returns the value for the given attribute name
      # Parameters
      # ==========
      #   package : DevopsLibrary::Providers::Package  
      #   attribute_name : The name of the attribute for which to get the value
      # Output
      # ======
      #   value for the attribute or nil if the attribute does not
      #   exist or does not have a value.
      #
      def get_attribute_by_name(package, attribute_name)
        value = nil
        attributes = package.data["ramAttributeValues"]
        attributes.each do |attribute|
          if attribute["attribute"]["value"]["name"] == attribute_name
            value = attribute["value"][0]
            break
          end
        end
        return value
      end

      # Create an asset and publish the artifacts to it.
      # Parameters
      # ==========
      #   package_id      : GUID of the asset
      #   package_version : Version of the asset
      #   packages        : Hash containing the files to be published to the asset
      #                     fully qualified file name => file name
      #   metadata        : Additional metadata that can be passed to create the asset
      #       build_url       : URL of the build
      #       name            : Application name
      #       description     : description of the asset to be created. 
      #                         If description is not passed, name would be set as description
      #       community_name  : Name of the community
      #       asset_type      : Asset Type of the asset
      #       attributes      : array of name / value
      #                         attribute_name : name of the attribute
      #                         attribute_value : value of the attribute
      #       classifications : classifications for the asset
      #   checksum         : If supplied, the aggregate checksum of all files in
      #                      packages hash
      # Output
      # ======
      #    Package : DevopsLibrary::Providers::Package 
      def create_publish_package(package_id, package_version, packages, metadata, checksum=nil)
        
        # Create the asset
        package = internal_create_package(package_id, package_version, metadata)

        # Publish each artifact to the asset (if there are any)
        if !packages.nil? && packages.size > 0
          
          # Add each artifact to the asset
          packages.each do|file,filename|
            publish_package(package, file, filename)
          end

          # If a checksum was not passed in, calculate it
          if checksum.nil?
            aggregate_checksum = calculate_aggregate_checksum(packages)
          else
            aggregate_checksum = checksum
          end

          # Add the Aggregate Checksum attribute to the asset
          add_attribute(package, "Build Checksum",  aggregate_checksum)
        end
          
        # Return the package
        return package
      end
      
      # Creates RAM asset.
      #  
      # Parameters
      # ==========
      #   package_id : GUID of the asset
      #   package_version : Version of the asset
      #   metadata : Additional metadata that can be passed to create the asset
      #        build_url : URL of the build
      #        name : Application name
      #        description : description of the asset to be created. 
      #                      If description is not passed, name would be set as description
      #        community_name : Name of the community
      #        asset_type : Asset Type of the asset
      #        attributes : array of name / value
      #                attribute_name : name of the attribute
      #                attribute_value : value of the attribute
      #        classifications : classifications for the asset
      #   
      # Output
      # ======
      #    Package : DevopsLibrary::Providers::Package 
      #
      #           
      #
      def internal_create_package(package_id, package_version, metadata={})
        # Don't seem to need root_folder
        response = nil
        
        package_id = handle_spaces(package_id)
        response = nil
         
        response = resolve_package(package_id, package_version,nil) if (!package_id.nil? && !package_version.nil?)  
       
        if response == nil 
            # Merge community and asset type in the metadata map
            asset_identification = {
              :id => package_id,
              :version => package_version 
            } 
            
            # Get the template file
            if package_id.nil? 
              template_file = File.join(File.dirname(__FILE__), 'createAssetTemplate.json')
            else
              template_file = File.join(File.dirname(__FILE__), 'createAssetTemplateGUID.json')
            end
            
            # Bind the template
            template = File.read(template_file)
            bindings = create_hash_from_json(asset_identification)
            asset_identification_as_json = evaluate(template, bindings) 
            asset_identification_hash = create_hash_from_json(asset_identification_as_json) 
            asset_metadata = metadata.merge(asset_identification_hash)
            
            #validate metadata 
            validate_metadata(asset_metadata)   
            
            #resolve relationship
            resolve_relationship(asset_metadata)
            
            #resolve categories
            resolve_categories(asset_metadata)
            
            #add build information
            add_build_attributes(asset_metadata)

            #add build category
            add_build_category(asset_metadata)
            
            #set contributor
            set_contributor(asset_metadata)

            # create Asset, passing template_file 
            data =  @ram_session.create_asset(asset_metadata.to_json) 
            idVal = data['id'] 
            url = get_url(idVal['ras:GUID'], idVal['ras:version']) 
            return DevopsLibrary::Providers::Package.new(idVal['ras:GUID'], idVal['ras:version'], data, url) 
            
        else
           url = get_url(package_id, package_version)
           return DevopsLibrary::Providers::Package.new(package_id, package_version, metadata, url)  
        end     
        
      end
      
      def set_contributor(asset_metadata)
          if !@build.nil? and @build.type=="Personal"
             user = @ram_session.get_user(@build.build_requestor_id)
             if !user.nil?
                  contributor = {
                     "resource" => user['href']
                  }
                  contributors = Array.new
                  contributors.push(contributor)
                  asset_metadata ["contributor"] = contributors
             end
          end
      end

    end
  end
end # end module DevopsLibrary 
      