require 'weaver/core/late_binding'

# Weaver-specific additions to Ruby's String class
class String
  # Return a +true+ or +false+
  def to_b
    return true if self == true || self =~ (/(true|t|yes|y|1)$/i)
    return false if self == false || self == '' || self =~ (/(false|f|no|n|0)$/i)
    raise ArgumentError.new("invalid value for Boolean: \"#{self}\"")
  end

  # Return a String replacing \:\: and \- with underscores
  def underscore
    self.gsub(/::/, '/').
    gsub(/([A-Z]+)([A-Z][a-z])/,'\1_\2').
    gsub(/([a-z\d])([A-Z])/,'\1_\2').
    tr("-", "_").downcase
  end

  # Return a symbol, with a similar value as the String, that can be used as a Weaver ID or Ruby identifier
  def symbolize
    # The original version did not always produce a symbol that was suitable as a Ruby
    # variable name.  rubyize_id makes sure the String is a valid symbol that can be expressed
    # as a Ruby variable, not a constant and not containing any non-ASCII characters
    #    return self.gsub(/\s+|-|\.|\(|\)/, "_").downcase.to_sym
    # Downcase to match the old behavior of String.symbolize (we don't actually need to do this to get a good Ruby identifier)
    str = to_s.downcase

    retval = (str.match /^[a-z_]/) ? '' : '_'   # prepend a _ if the initial character wouldn't make a good variable name initial character

    str.each_char do |c|
      # If we get an unexpected ID character, use the ASCII representation
      c = '_' if c.match(/\s+|-|\.|\(|\)/) # Don't use ASCII representations of -, ., (, ), or ' '
      retval = retval + ((c.match /[a-zA-Z0-9_]/) ? c : ('_' + c.unpack('C')[0].to_s))
    end

    return retval.to_sym
  end

  # @private
  alias_method :concat_original, :concat

  # Support {Weaver::Core::LateBinding}
  def concat(obj)
    if obj.kind_of?(Weaver::Core::LateBinding)
      # Late Bindings are evaluated using SandBox.instance_eval()
      # TODO escape 'self' so that it can safely appear in single quotes.
      code = "Proc.new { |late_binding| #{dump} + late_binding.weaver_eval(@entity, @dependencies) }"
      return Weaver::Core::LateBinding.new(nil, obj.declaring_scope, obj, &(eval(code)))
    end
    concat_original(obj)
  end

  alias_method :plus_original, :+

  # Support Weaver::Core::LateBinding
  def +(obj)
    if obj.kind_of?(Weaver::Core::LateBinding)
      # Late Bindings are evaluated using SandBox.instance_eval()
      # TODO escape 'self' so that it can appear in quotes.
      code = "Proc.new { |late_binding| #{dump} + late_binding.weaver_eval(@entity, @dependencies) }"
      return Weaver::Core::LateBinding.new(nil, obj.declaring_scope, obj, &(eval(code)))
    end
    plus_original(obj)
  end

  # Ensure that we can turn off colorization, bold and italic
  @@colorize = true
  
  def self.colorize=(colorize)
    @@colorize = colorize
  end
  
  def self.colorize?
    @@colorize
  end
  
  # colorization
  def colorize(color_code)
    if @@colorize
      "\e[#{color_code}m#{self}\e[0m"
    else
      self
    end 
  end

  def red
    colorize(31)
  end

  def green
    colorize(32)
  end

  def yellow
    colorize(33)
  end
  
  def blue
    colorize(34)
  end

  def pink
    colorize(35)
  end
  
  def grey
    colorize(37)
  end
  
  def bold
    colorize(1)
  end

end