require 'weaver/core/base'
require 'weaver/utils/reversing_hash'
require 'weaver/core/group_container_mixin'
require 'weaver/core/generation'

module Weaver
  module Core
    # Group represents a collection of Weaver Base entities and some attributes about them,
    # for example the members of a WAS cell and the cell name they share.
    #
    # Author:: Ed Snible
    class Group < Base
      include GroupContainerMixin            # Group can include other Groups
      include Generation
      
      # Hash of :id to Base
      attr_accessor :members
      
      # * +id+ Symbol to represent the entity
      # * +parent+ The enclosing Base
      def initialize(id, parent)
        super(id, parent)
        
        @type = :group
        
        @members = Weaver::Utils::ReversingHash.new(self, :member_of)
      end
      
      def initialize_copy(other)
        super
        
        # Clear out the shallow copies that members is pointing to, we will re-link it later.
        member_ids = members.keys
        @members = Weaver::Utils::ReversingHash.new(self, :member_of)
        member_ids.each { |member_id| @members[member_id] = nil }
        
        initialize_copy_groups(other)
      end
      
      # :stopdoc:
      def method_missing(sym, *arguments, &block)
        # allow members to be referenced by name
        return @members[sym] if @members.has_key?(sym)
        
        super
      end
      # :startdoc:
      
      # Validates the entity and its direct and indirect contained children.
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_collect(out_messages, desired_phases = nil)
        @@logger.debug "Validating group '#{id}'"
        @@logger.debug "  has #{members} members"
        
#        freshen_members
        
        super
        
        validate_groups(out_messages, desired_phases)
        
        @members.each do |id, member|
          out_messages.push ProblemStatus.new(Logger::ERROR, "Unknown member :#{id}", self) if member.nil?
        end
      end
      
#      # Refresh @members
#      #
#      # :stopdoc
#      # (This is the old-style of freshening, we should replace this with register_reference_freshener()
#      # which works even without validation being invoked.)
#      # :startdoc
#      def freshen_members
#        @members.each do |id, member|
#          if member.nil?
#            # Note that :was_node1 is the parent.child.child.child and scoped_lookup() will not find it.
#            # With groups we don't have any expectation the id will be in the groups lexical scope
#            # We could have the user write the full expression, but this would suck as we can't evaluate
#            # it until we have parsed a little further than we are now, and it is also a lot of IDs
#            # to get write.  Instead we will look globally after parsing has finished (which traversal is best?)
#            member = root_entity.flatten_children.find { |entity| entity.id == id }
#            unless member.nil?
#              members[id] = member
#            end
#          end
#        end
#      end
      
      # Returns true if this Group is equal to another Group
      def ==(other)
        retval = super
        return retval && @members == other.members
      end
      
      # Return a String containing a Weaver description of the group members
      def to_weaver_group_members
        # derive template name according to convention
        template_file = 'data/templates/weaver_core_group_members.erb'
        file = File.join(File.dirname(__FILE__), "..", "..", "..", template_file)
        unless File.exists?(file)
          raise WeaverError, "Cannot find Erubis template file #{template_file} for Group memberes."
        end
        return Weaver::Utils::Indent.fixup(Erubis::Eruby.new(File.read(file)).result(binding))
      end
      
    end
    
  end
end