require 'net/scp'
require 'net/ssh'
require 'ostruct'
require 'rubygems'
require 'socket'

require 'weaver/core/base.rb'

module Weaver
  module Core
    
    # A LoadDriver that generates load for a given component.
    # 
    # Author:: Thomas Gschwind <thg@zurich.ibm.com>
    # Copyright:: IBM
    class LoadDriver < Base
      include ImportContainerMixin
      include PolicyContainerMixin
      
      # The assertions that should be met
      attr_accessor :asserts
      
      def create_attribute(sym)
        m_attribute = Attribute.new(sym, self, true)
        m_attribute.implicit = true
        m_attribute.required = false
        amd[sym] = m_attribute
        
        return m_attribute        
      end
      
      def initialize(id, weaver_parser)
        super(id, weaver_parser)
        @type = :load_driver
        @asserts = []
        
        # Description of the host from where load is driven
        create_attribute(:driver_hostname)
        create_attribute(:driver_username)
        create_attribute(:driver_password)
        create_attribute(:driver_command)
        
        # Local and remote data directory where information is stored
        create_attribute(:local_data_dir)
        create_attribute(:remote_data_dir)
        
        # Description of the host load should be driven to - set automatically
        create_attribute(:target_host)
        create_attribute(:target_port)
      end
      
      def validate_collect(out_messages, desired_phases = nil)
        @@logger.debug "Validating load driver '#{id}'"
        super
      end
      
      # Evaluate the assert statements attached to this load driver, the
      # results will be added to the out_messages list.
      def evaluate_asserts(ld_data, out_messages)
        data = OpenStruct.new
        ld_data.each do |key, val|
          data.send("#{key}=", val)
        end

        res = true
        @asserts.each do |assert|
          res2 = assert[1].call(data)
          out_messages.push [assert[0], res2] 
          res &= res2
        end
        return res
      end
      
      # Evaluate the assert statements attached to this load driver and
      # write the result to junit_filename as a junit xml file.
      def to_junit(ld_data, junit_filename)
        tests = []
        res = self.evaluate_asserts(ld_data, tests)
          
        # generate junit xml
        xml = ""
        xerrs = 0
        xtests = 0
        tests.each do |e|
          xml += "    <testcase classname='#{e[0].to_s}' name='#{e[0].to_s}'"
          if e[1]
            xml += "/>\n"
          else
            xml += ">\n"
            xml += "      <failure  message='Assertion failed, please check JMeter report for details' type='failed'/>\n"
            xml += "    </testcase>\n"
            xerrs += 1
          end
          xtests += 1
        end
    
        x = "<?xml version='1.0' encoding='UTF-8'?>\n" +
            "<testsuite failures='#{xerrs.to_s}' errors='#{xerrs.to_s}' name='JMeter JKE Web Load Testing' tests='#{xtests.to_s}' time='#{ld_data['driver_duration'].to_f}'>\n" +
              xml.to_s + 
            "</testsuite>\n"
          
        status = "OK"
        if xerrs > 0
          status = "ERROR"
        end

        # write the data into a file
        File.open(junit_filename, 'w') { |f| f.write(x) }
        
        return status
      end

      # Evaluate the assert statements attached to this load driver and
      # write the result to junit_filename as a junit xml file.
      def to_itm_junit(policy, policy_data, junit_filename)
        x = "<?xml version='1.0' encoding='UTF-8'?>\n" +
            "<testsuite failures='#{xerrs.to_s}' errors='#{xerrs.to_s}' name='JMeter JKE Web Load Testing' tests='#{xtests.to_s}' time='#{ld_data['driver_duration'].to_f}'>\n" +
            "  <testcase classname='#{policy}' name='#{policy}'>\n"+
            "    <failure message='Policy triggered #{policy_data.length} times, check corresponding json file for details' type='failed'/>\n"
            "  </testcase>\n"
            "</testsuite>\n"          
        # write the data into a file
        File.open(junit_filename, 'w') { |f| f.write(x) }
        return "ERROR"
      end

      # TODO: fix the signature of this method
      def generate_load(build_system, build)
        tmpdir = '/tmp'
        target_host = self.target_host
        target_port = self.target_port

#        # Find parent component and node - disabled until late_binding is in place
#        load_driver = self
#        component = nil
#        node = nil
#        n = self
#        while !n.nil? do
#          puts "#{n}"
#          if n.kind_of? ComponentStatement 
#            component = n
#          elsif n.kind_of? NodeStatement
#            node = n
#          end
#          n = n.scope
#        end
          
#        # TODO: only for testing in Windows 7 without deploying first        
#        if ENV["TOM_DEBUG"] == "yes"
#          tmpdir = 'C:/cygwin/home/thg/tmp'
#          target_host = '9.37.129.197'
#          target_port = 8080
#        end
        
        if build.nil?
          build_id = 'xxx'
        else
          build_id = build.object_id
        end
        
        puts "ld(#{self.id}): started"
        d_hostname = self.driver_hostname
        d_username = self.driver_username
        d_password = self.driver_password
        l_dir = self.local_data_dir #'jmeter'
        r_dir = self.remote_data_dir #'jmeter'
        r_input_dir = File.join(r_dir, 'input', self.id.to_s)
        r_output_dir = File.join(r_dir, 'output', self.id.to_s)
        l_output_dir = File.join(tmpdir, "output_#{build_id}")
        FileUtils.mkpath l_output_dir
        
#        # TODO: shall we upload a user provided data directory, if so, do it here
#        puts "ld(#{self.id}): h=#{d_hostname} u=#{d_username} p=#{d_password}"
#        Net::SCP.start(d_hostname, d_username, :password => d_password, :paranoid => false) do |scp|
#          scp.upload!("@@@/#{l_dir}", r_input_dir, :recursive => true)
#        end
        
        ################################
        # Iterate over all policies attached to the current load driver and enable them
        self.policies.each { |key, pol|
          result = pol.service 'start'
          puts "ld(#{self.id}): start #{key} => #{result}"
        }
        
        # TODO: the directory from which the tests are executed from are currently hardcoded
        puts "ld(#{self.id}): generating load"
        # TODO: maybe the cmd line should be more user extensible with any type of infrastructure parameters
        ld_start = Time.new
        Net::SSH.start(d_hostname, d_username, :password => d_password, :paranoid => false) do |ssh|
          ssh.exec! "#{self.driver_command} #{self.id} #{target_host} #{target_port}"
        end
        ld_duration = Time.new - ld_start
        
        ################################
        # Iterate over all policies attached to the current load driver,
        # disbale them and collect any external links
        self.policies.each { |key, pol|
          result = pol.service 'stop'
          puts "ld(#{self.id}): stop #{key} => #{result}"
          
          unless result.nil?
            result_file = File.join(l_output_dir, "#{self.id}_#{key}.json")
            File.open(result_file, 'w') { |f| f.write(result) }
            build_system.publish_log_file(build, "#{self.id}_#{key}.json", result_file, 'OK')
            
            begin
              json_result = JSON.parse(result)
              json_result.each { |jkey, jval|
                if jkey =~ /_url$/
                  build_system.publish_link(build, jval, "#{jkey.sub(/_url$/, '')}: #{self.id}_#{key}") unless build_system.nil?
                end
              }
              if pol.type == :itm_policy
                FileUtils.mkpath File.join(l_output_dir, 'tests')
                junit_filename = File.join(l_output_dir, 'tests', "#{key}_junit.xml")
                self.to_itm_junit(pol, json_result, junit_filename)
              end
            rescue
              puts "Cannot parse JSON result #{result}"
            end
          end
          puts "xxx#{result}xxx"
        }
        
        ################################################################
        # now that the policies are stopped, obtain any data generated from
        # the load driver and post it inside the RTC build
        Net::SCP.start(d_hostname, d_username, :password => d_password, :paranoid => false) do |scp|
          scp.download!(r_output_dir, l_output_dir, :recursive => true)
        end
        
        # check if results fulfill the assertions defined as part of the
        # load driver
        l_output_dir = File.join(l_output_dir, self.id.to_s)
        sla_files = Dir.glob(File.join(l_output_dir, 'logs', '*.sla.json'))
        sla_status = "OK"
        if sla_files.length>0
          if sla_files.length>1 
            puts "More than 1 sla.json file, using only first one"
          end
          result = JSON.parse(File.read(sla_files[0]))
          result['driver_duration'] = ld_duration 
          puts "error_rate: #{result['error_rate']}"
          puts "response_time_avg: #{result['response_time_avg']}"
          puts "response_time_max: #{result['response_time_max']}"
        
          FileUtils.mkpath File.join(l_output_dir, 'tests')
          junit_filename = File.join(l_output_dir, 'tests', 'junit.xml')
          sla_status = self.to_junit(result, junit_filename)
        end

        # publish files to build
        puts "p=#{File.join(l_output_dir, 'logs', '*')}"
        log_files = Dir.glob(File.join(l_output_dir, 'logs', '*'))
        puts "lf = #{log_files}"
        log_files.each { |f|
          puts "#{f} -> logs"
          build_system.publish_log_file(build, "#{f}", f, 'OK') unless build_system.nil?
        }
        
        link_files = Dir.glob(File.join(l_output_dir, 'links', '*'))
        puts "lf = #{link_files}"
        link_files.each { |f|
          link = IO.read(f)
          puts "#{f} #{link} -> links"
          build_system.publish_link(build, link, "#{f}") unless build_system.nil?
        }
        
        test_files = Dir.glob(File.join(l_output_dir, 'tests', '*'))
        puts "tf = #{test_files}"
        test_files.each { |f|
          puts "#{f} -> tests"
          build_system.publish_junit_results(build, f) unless build_system.nil?
        }
        
        puts "ld(#{self.id}): done"
        return sla_status == "OK" 
      end

    end
  end
end
