require 'weaver/core/logging'
require 'weaver/transforms/fold'
require 'weaver/transforms/duplicate'
require 'string'

module Weaver
  module Core

    # Implements Weaver transformations such as 'topology folding' and also provides a duplicate() method
    # for Weaver entities (currently only for Environment)
    #
    # Author:: Ed Snible
    module WeaverTransforms
      include Weaver::Core
    
      # Add everything in in_env2 to in_out_env.  This means infrastructure and application nodes
      # and links.  in_env2 will not be modified.    
      def self.merge_environments(in_out_env, in_env2)
        env2 = Weaver::Transforms::Duplicate.duplicate(in_env2)
        
        if in_out_env.infrastructure.nil?
          in_out_env.infrastructure = env2.infrastructure
        else
          merge_infrastructures(in_out_env.infrastructure, env2.infrastructure) unless env2.infrastructure.nil?
        end

        env2.topologies.each do |id, top|
          if in_out_env.topologies[id].nil?
            in_out_env.topologies[id] = top
          else
            merge_topologies(in_out_env.topologies[id], top)
          end
        end
        
        # TODO startup order
      end
      
      # Improve the IDs by using names    
      #
      # * *Args*    :
      #   - +env_or_inf+ -> a Topology
      def self.improve_ids(env_or_inf)
        raise "Paramter 'env_or_inf' must be an Environment or Infrastructure but is a '#{env_or_inf.type}'" unless env_or_inf.kind_of?(Topology) 

        # handle environment
        if env_or_inf.respond_to?(:topologies)
          env = env_or_inf
          env.topologies.each do |id, topology|
            raise WeaverError.new("#{env.type} (#{env.id}) has nil topology (#{id})") if topology.nil?        
            newnodes = {}
            topology.nodes.each { |id, node| improv_node_ids(node, newnodes.merge(topology.nodes)); newnodes[node.id] = node }
            topology.nodes.clear
            topology.nodes.merge!(newnodes)
          end
        end
        
        # handle infrastructure
        if env_or_inf.kind_of?(Topology)
          inf = env_or_inf
          newnodes = {}
          inf.nodes.each { |id, node| improv_node_ids(node, newnodes.merge(inf.nodes)); newnodes[node.id] = node }
          inf.nodes.clear
          inf.nodes.merge!(newnodes)
        end        
      end
      
      # Improve the IDs by replacing those that are not valid Ruby identifiers with valid Ruby identifiers.
      # For example :"heiz\303" becomes :heizo_303.
      #
      # * *Args*    :
      #   - +base+ -> a Base
      def self.rubyize_ids(base)
        base.id = rubyize_id(base.id)
        base.flatten_children.each { |child| child.id = rubyize_id(child.id) }
      end
      
      # Given a Ruby Symbol or String, return a similar Ruby Symbol that can be represented without quotes.
      # DEPRECATED Use String.symbolize
      #
      # * *Args*    :
      #   - +sym_or_string+ -> a Symbol or String
      # * *Returns* :
      #   - Symbol
      def self.rubyize_id(sym)
        return sym.to_s.symbolize
      end
      
      # nodenames is the set of names -- current and future -- that we need to avoid collisions with
      private
      def self.improv_node_ids(node, nodenames)
        symb = node.name.symbolize unless node.name.nil?
        node.id = symb unless nodenames.key?(symb) 
        newcomponents = {}
        node.components.each do |id, component|
          unless component.name.nil? 
            symb = component.name.symbolize 
            component.id = symb unless nodenames.key?(symb)
          end
          newcomponents[component.id] = component
        end
        node.components = newcomponents 
      end
      
      def self.merge_infrastructures(in_out_inf, inf2)
        in_out_inf.validators.concat(inf2.validators)
        
        # We ignore attribute collisions
#        in_out_inf.unresolved_attributes.merge!(inf2.unresolved_attributes)
        in_out_inf.amd.merge!(inf2.amd)
        
        # TODO if the clouds are different push inf2's cloud into all inf2.nodes
        
        in_out_inf.nodes.merge!(inf2.nodes)
      end
      
      def self.merge_topologies(in_out_top, top2)
        in_out_top.validators.concat(top2.validators)
        
        # We ignore attribute collisions
        in_out_top.amd.merge!(top2.amd)
        
        in_out_top.nodes.merge!(top2.nodes)
        in_out_top.links.merge!(top2.nodes)
        
        # TODO Check for username collision?
        in_out_top.users.merge!(top2.nodes)
        
        in_out_top.topologies.merge!(top2.topologies)
      end
      
    end
  end
end
