require 'socket'

module Weaver
  module Deployers
    module Execution   
      # This class takes care of setting up the nodes of a root model so that Weaver can automatically learn
      # continuous validation rules by observing how each piece of automation changes the runtime state
      #
      # Author:: Fabio Oliveira
      class AnalyticsSetup
        include Weaver::Core
        # +root+: Weaver root model
        # +weaver_project_dir+: Directory of the current Weaver project
        # +service_port+: Port on which the Weaver service listens
        def initialize(root, weaver_project_dir, service_port = 9999)
          @root = root
          @weaver_project_dir = weaver_project_dir
          @weaver_service_port = service_port
        end
        
        # Adds to a node a Chef automation element corresponding to the 'analytics' recipe
        # This Chef automation is structured as follows:
        #     chef_automation(:weaver_analytics) { 
        #       role "weaver_analytics"
        #       property :weaver_analytics_service_hostname, :name => "weaver.analytics.service_hostname"
        #       property :weaver_analytics_service_port, :name => "weaver.analytics.service_port"
        #       property :weaver_analytics_weaver_project_dir, :name => "weaver.analytics.weaver_project_dir"
        #       property :weaver_analytics_deployment_id, :name => "weaver.analytics.deployment_id"
        #       property :weaver_analytics_fq_node_id, :name => "weaver.analytics.fq_node_id"
        #       property :weaver_analytics_node_id, :name => "weaver.analytics.node_id"
        #     } 
        def add_analytics_automation(deployment_id)
          @root.flatten_children.select {|c| c.weaver_instance_of?(:node) && c.realized_by.empty? && !Weaver::Transforms::Unroll.template?(c)}.each do |node|
            @@logger.debug "Adding analytics automation to node '#{node.id}'"
            
            automation = Weaver::Core::Automation.new(:weaver_analytics, node)
            automation.type = :chef_automation

            role = Weaver::Core::Attribute.new(:role, automation)
            role.value = "weaver_analytics"
            automation.amd[:role] = role

            deps_file_url = Weaver::Core::Attribute.new(:deps_file_url, automation)
            deps_file_url.name = "DEPS_FILE_URL"
            automation.amd[:deps_file_url] = deps_file_url

            attribute = Weaver::Core::Attribute.new(:weaver_analytics_service_hostname, automation)
            attribute.name = ("weaver.analytics.service_hostname")
            attribute.value = Socket.ip_address_list.detect{|i| i.ipv4? && !i.ipv4_loopback? && !i.ipv4_multicast? && !i.ipv4_private?}.ip_address
            attribute.mutable = true
            automation.amd[attribute.id] = attribute

            attribute = Weaver::Core::Attribute.new(:weaver_analytics_service_port, automation)
            attribute.name = ("weaver.analytics.service_port")
            attribute.value = @weaver_service_port
            attribute.mutable = true
            automation.amd[attribute.id] = attribute
            
            attribute = Weaver::Core::Attribute.new(:weaver_analytics_weaver_project_dir, automation)
            attribute.name = ("weaver.analytics.weaver_project_dir")
            attribute.value = @weaver_project_dir
            attribute.mutable = true
            automation.amd[attribute.id] = attribute
            
            attribute = Weaver::Core::Attribute.new(:weaver_analytics_deployment_id, automation)
            attribute.name = ("weaver.analytics.deployment_id")
            attribute.value = deployment_id
            attribute.mutable = true
            automation.amd[attribute.id] = attribute            

            attribute = Weaver::Core::Attribute.new(:weaver_analytics_fq_node_id, automation)
            attribute.name = ("weaver.analytics.fq_node_id")
            attribute.value = node.fully_qualified_id
            attribute.mutable = true
            automation.amd[attribute.id] = attribute
            
            attribute = Weaver::Core::Attribute.new(:weaver_analytics_node_id, automation)
            attribute.name = ("weaver.analytics.node_id")
            attribute.value = node.id
            attribute.mutable = true
            automation.amd[attribute.id] = attribute

            node.automation[automation.id] = automation
            node.automation_order.insert(0, automation)
          end               
        end
      end
    end
  end
end