require 'weaver/core/logging'
require 'weaver/dsl/requires/type_handler'
require 'weaver/core/requirements'

module Weaver
  module Dsl
    module Requires
      # Author:: Fabio Oliveira
      class StorageTypeHandler < TypeHandler
        def initialize()
          super [ :storage ]
        end
       
        def handle(parent, args)
          @@logger.debug "#{self.class} #{parent} #{args}" 
          return StorageRequirement.new(parent.model, args) 
        end
      end
      
      class StorageRequirement < Core::Requirement
        attr_accessor :required_size
        attr_accessor :required_unit
        
        def initialize(scope, requires_hash)
          super scope
        
          @name = requires_hash[:storage]
          @required_size = requires_hash[:size]
          @required_unit = requires_hash[:unit]
          # TODO warn if we have any additional keys in requires_hash besides :description
        end
        
        def validate_collect(out_messages, desired_phases = nil)
          #inf_node = get_final_realization()
          all_final_realizations.each do |inf_node|
            unless inf_node.nil?
              @@logger.debug "Validating storage requirement. Node = #{inf_node.id}"
              
              storage = find_target_storage(inf_node)
              if storage.nil?
                @@logger.debug "Could not find target storage for node #{inf_node.id}"
                out_messages.push ProblemStatus.new(Logger::ERROR, generate_storagenotattached_message(inf_node), inf_node, weaver_origin)
              else
                @@logger.debug "Found node's target storage: #{storage.inspect}"
                required_size = convert_to_MB(@required_size, @required_unit)
                target_size = convert_to_MB(storage.size, storage.unit)
                if required_size > target_size
                  out_messages.push ProblemStatus.new(Logger::ERROR, generate_insufficientstorage_message(inf_node, storage), inf_node, weaver_origin)
                end
              end
            end
          end
        end
        
        private
        
        def convert_to_MB(size, unit)
           if unit == 'MB'
             return size
           elsif unit == 'KB'
             return size / 1024
           elsif unit == 'GB'
             return size * 1024
           end
        end
        
        def find_target_storage(inf_node)
          return nil if inf_node.nil?
          
          current = inf_node.scope
          while !current.nil?
            @@logger.debug "Processing ancestor: #{current.id}"
            current.flatten_children.each do |child|
              if child.weaver_instance_of?(:storage_link)
                if child.source.id == inf_node.id
                  return child.target
                end
              end
            end
            current = current.scope 
          end
          return nil
        end
        
        def generate_storagenotattached_message(target)
             return <<-eos
        #{weaver_classname(@scope)} :#{@scope.id} at (#{@weaver_origin}) requires the
        #{weaver_classname(target)} :#{target.id} at (#{target.weaver_origin}) to be attached to a storage volume of size '#{@required_size} #{@required_unit}', 
        but it was not attached to any storage'
        eos
        end
        
        def generate_insufficientstorage_message(target, target_storage)
          " --> #{weaver_classname(@scope)} :#{@scope.id} at (#{@weaver_origin}) requires a storage volume of size '#{@required_size} #{@required_unit}', but #{weaver_classname(target)} :#{target.id} at (#{target.weaver_origin}) was attached to a smaller volume of size '#{target_storage.size} #{target_storage.unit}'" 
        end
      end
    end
  end
end