require 'singleton'
require 'weaver/core/logging'

# include all handlers
Dir[File.dirname(__FILE__) + "/*_type_handler.rb"].each do |file|
  require file
end

module Weaver
  module Dsl
    module Requires      
      
      # Simple implementation of a TypeHandlerRegistry. Loads all type handlers for 
      # the Weaver 'requires' statement.
      # *Example:*
      # requires type => :package, :name => 'mysql-server'
      #
      class TypeHandlerRegistry
        include Singleton
        include Weaver::Core

        # A hash of all TypeHandler implementations. 
        # Key:: Symbol reflecting the type that is handled, e.g., :package, :os_service, etc
        # Value:: An implementation of the TypeHandler that can handle the type specified as the key.  
        attr_reader :handlers
        
        def initialize
          @handlers = Hash.new
          # load  all *TypeHandler file name to derive class name
          Dir[File.dirname(__FILE__) + "/*_type_handler.rb"].each do |file|
            class_name = camelize(File.basename(file)[0..-4]);

            module_inst = Module.const_get("Weaver").const_get("Dsl").const_get("Requires")            
            handler_class = module_inst.const_get class_name
            # initialize handler
            handler = handler_class.new

            # check if it has a method "handles"
            if (!handler.respond_to?("handle"))
              error_msg = "Error: #{handler_class} does not define a 'handles' method.";
              @@logger.fatal msg
              raise WeaverError.new(msg)
            end

            handler.handles.each { |item|
              @handlers[item] = handler
              @@logger.debug "Loading #{handler} for type: #{item}"
            }
          end

        end

        private

        def camelize(str)
          str.split('_').map {|w| w.capitalize}.join
        end

        public

        def find(type)
          return @handlers[type]
        end

      end
    end
  end
end