require 'timeout'
require 'weaver/core/logging'
require 'weaver/core/topology'
require 'weaver/core/weaver_error'
require 'weaver/deployers/execution/coordination_util'
require 'weaver/deployers/execution/storage_modifier'
require 'weaver/execution/actions/action'
require 'weaver/providers/delegating_provider'
require 'weaver/system'

module Weaver
  module Execution
    module Actions
    
      # Runs a set of automations on a set of nodes
      
      class RunAutomationsAction < Weaver::Execution::Actions::Action
        
        DEFAULT_TOTAL_TIME = 600 # seconds
        DEFAULT_WAIT_POLLS_FOR_STARTUP = 160
        
        # operation_map is a Hash mapping node to an array of operations (automations)
        def initialize (operation_map = {}, execute_parameters = {})
          @operation_map = operation_map
          @execute_parameters = execute_parameters
          
          @deployers = {}
        end # end def initialize

        # Execute the automations on the nodes as specified in @operation_map
        def execute (properties, execute_parameters = {})
          execute_parameters.each {|key,value| @@logger.debug "execute called with '#{key}' => #{value}" }
          return if @operation_map.empty?
          raise Weaver::Core::WeaverError.new("Coordinator required") if !use_coordination
          raise Weaver::Core::WeaverError.new("Coordination transaction identifier required.") if execute_parameters[:coordination_id].nil?
          
          new_nodes = []
          updated_nodes = []

          # compute dependencies and add extra parameters
          dependencies = Weaver::Deployers::Execution::CoordinationUtil.get_dependencies(@operation_map.keys.first.root_entity)
          execute_parameters.merge!(:weaver_coordination_topology_id => plan.transaction.id)
          execute_parameters.merge!(:weaver_coordination_dependencies => dependencies)
          
          # Some of the source values for dependencies are in the model. Write them to the coordinator now.
          Weaver::Deployers::Execution::CoordinationUtil.write_constants(plan.transaction, dependencies)

          # Identify deployers for each node
          # create Hashes Weaver::Deployers::Deployer :-> Array<Weaver::Core::Node>
#          nodes_to_update_by_deployer = {}
#          nodes_to_create_by_deployer = {}
          nodes_by_deployer = {}
          @operation_map.each do |node, automations|
            provisioner = node.provisioner || Weaver::Providers::DelegatingProvider.new()
            deployer = provisioner.deployer(properties, &@reporter)
            @@logger.debug "Identified deployer #{deployer} for node #{node.fully_qualified_id}"
            @deployers[node] = deployer
            nodes_by_deployer[deployer.class.id] = [*nodes_by_deployer[deployer.class.id]] << node
#            if deployed?(node) then
#              nodes_to_update_by_deployer[deployer] = [*nodes_to_update_by_deployer[deployer]] << node
#            else
#              nodes_to_create_by_deployer[deployer] = [*nodes_to_create_by_deployer[deployer]] << node
#            end # if 
          end # @operation_map.each do
          
          # deploy nodes by deployer; this implementation assumes coordination will deal w/ order
#          nodes_to_create_by_deployer.each do |deployer, nodes|
          nodes_by_deployer.each do |deployer, nodes|
          @@logger.debug %Q{Calling #{deployer}.update with:\n#{nodes.map{|node| node.fully_qualified_id}.join("\n")}}
            @deployers[nodes.first].update(nodes, @operation_map, execute_parameters)
#            if deployer.respond_to?(:deploy_pattern) then
#              deployer.deploy_nodes(nodes, parameters)
#            else
#              nodes.each do |node| 
#                @@logger.debug "Deploying new VM for node #{node.fully_qualified_id}"
#                deployer.deploy_node(node, execute_parameters)
#                new_nodes << node
#              end # nodes.each do
#            end # if deployer.respond_to?(:deploy_pattern)
          end # @nodes_to_create_by_deployer.each do

#          nodes_to_update_by_deployer.each do |deployer, nodes|
#            nodes.each do |node|
#              @@logger.debug "Updating existing VM (#{node.ip_address}) for node #{node.fully_qualified_id}"
#              update_node(node, execute_parameters)
#              updated_nodes << node
#            end # nodes.each do
#          end # @nodes_to_create_by_deployer.each do
#
#          # TODO: the remaining actions should go done asynchronously
#          # wait for new nodes to start
#          wait_for_start (new_nodes.dup)
#          
#          # any final actions by the deployer
#          new_nodes.each { |node| @deployers[node].post_start_actions(node, execute_parameters) }
#          
          @@logger.debug "Completed execute"
        end # end def execute
                      
        # Determine whether or not coordination is being used
        # At this point, we pretty much assume that it is
        def use_coordination
          return !Weaver::System.instance.get_property(:zookeeper_hosts).nil?
        end

#        def locally_resolvable(dependencies) 
#          return dependencies.select { |d| 
#            Weaver::Deployers::Execution::CoordinationUtil.get_container(d[:source], Weaver::Core::Automation).nil? and !['hostname', 'ip_address'].member?(d[:source].id)
#          }
#        end
#        
#        # Write values to coordinator from the model. This should be done for attributes on which there are dependencies
#        # defined (because they are cross node) but for which the value is available in the model.
#        # *Args*
#        #  + dependencies + dependencies that are locally resolvable; assumed to be so 
#        def write_constants (dependencies)
#          dependencies.each do |dependency|
#            source = dependency[:source]
#            source_node = Weaver::Deployers::Execution::CoordinationUtil.get_container(source, Weaver::Core::Node)
#            # The path from the node to the attribute
#            suffix = source.fully_qualified_id.sub(%r{^#{source_node.fully_qualified_id}}, '')
#        
#            # All the nodes that need to be considered (group members)
#            nodes = Weaver::Transforms::Unroll.template?(source_node) ? Weaver::Transforms::Unroll.group(source_node).members.values : [source_node]
#            nodes.each do |node|
#              value = node.instance_eval(suffix.sub(/^\./, ''))
#              plan.transaction.write_data(node.fully_qualified_id + suffix, value.to_s) unless value.nil?
#            end # nodes.each do
#          end # dependencies.select {}.each do
#        end # def self.write_constants
#
        def to_s
          retval = "RunAutomationAction["
          retval << @operation_map.map { |node, automations| "   #{node.fully_qualified_id}[#{automations.map{|a| a.id.to_s}.join(',')}]" }.join(';')
          retval << "]"
          
          return retval
        end # def to_s

      end # class RunAutomationsAction
    
    end # module Action
  end # end module Execution
end # module Weaver