require 'weaver/core/automation'
require 'weaver/core/logging'
require 'weaver/core/node'
require 'weaver/deployers/execution/coordination_util'
require 'weaver/execution/actions/action'
require 'weaver/execution/actions/delete_nodes_action'
require 'weaver/execution/actions/run_automations_action'
require 'weaver/execution/actions/wait_action'
require 'weaver/transforms/merge'

module Weaver
  module Execution
    
    class ChangePlan
      include Weaver::Core
      attr_accessor :root
      attr_accessor :plan # TODO remove ?
      attr_accessor :change_template  #TODO remove ?
      attr_accessor :transaction
      
      def initialize(change_template, template_config, desired_state, current_state)
        @@logger.debug "Creating change plan for #{current_state.fully_qualified_id unless current_state.nil?} --> #{desired_state.fully_qualified_id} with request:"

        # define @change_template
        retrieve_template(change_template)
        specialize_template(template_config || {})
        
        # define @root
        @desired_state = desired_state
        @@logger.debug "@desired_state = #{@desired_state}"
        # the current_state and desired_state need to be merged
        @current_state = current_state
        @@logger.debug "@current_state = #{@current_state}"
        @root = Weaver::Transforms::Merge.merge(@desired_state,@current_state)
        
        # initialize @plan
        @action_index = 0
        @plan = []
        # The following will also define/use @assignments_block and @defaults_block
        self.instance_eval(@change_template)
      end
      
      # Handle overrides of default values for the variables in the action block.
      # An "assignments" block is typically generated in ChangeRequest#new
      def assignments (&block)
        @assignments_block = block
      end
      
      # Handle assignment of default values to variables in the action block.
      # A "defaults" block should be part of any template definition.
      def defaults (&block)
        @defaults_block = block 
      end
      
      # Handle the creation of the plan using values defined in the defaults block and overridden in the assignments block.
      def actions (&block)
        @defaults_block.call if defined?@defaults_block
        @assignments_block.call if defined?@assignments_block
        yield
      end

      # Define an action step that runs automations (w/ particular tags) on a set of nodes      
      def run_automations(automation_tags = [], nodes = [])
        automation_tags = [ :_default ] unless defined?(automation_tags)
        unless automation_tags.empty?
          @@logger.debug "Identifying automations with tags: #{automation_tags.inspect}"   
          add_action(Weaver::Execution::Actions::RunAutomationsAction.new(create_operations_map(@root, [*automation_tags], [*nodes], true)))
        end
      end # def run_automations
      
      # Define an action to add nodes to a deployed environment.
      # When executed, this action will add nodes to the environment IF the update environment (@root) indicates 
      # that the number of nodes with the specified identifiers is to increase.
      # * +nodes+ is a list of node identifiers
      def add_nodes(nodes = [])
        @@logger.debug "Adding nodes: #{[*nodes]}"
        add_action(Weaver::Execution::Actions::RunAutomationsAction.new(create_operations_map(@root, [ :_default ], [*nodes], false)))
      end # def add_nodes
      
      # Define an action to remove nodes from a deployed environment.
      # When executed, this action will remove nodes from the environment IF the update environment (@root) 
      # indicates that the number of nodes with the specified identifiers is to decrease.
      # * +nodes+ is a list of node identifiers
      def remove_nodes(nodes = [])
        @@logger.debug "Removing nodes: #{[*nodes]}; not fully supported"
        identify_nodes_to_delete([*nodes]).each do |node|
          add_action(Weaver::Execution::Actions::DeleteNodesAction.new([node]))
        end
      end
      
      # Define an action step to wait.  
      # A wait action can (a) specify that certain prior tasks should be completed before continuing and/or
      #                   (b) a fixed delay time
      def wait(automation_tags = [], nodes = [], time = 0)
        time = time.to_i if time.is_a?(String)
        @@logger.debug "Waiting #{time} ms"
        # TODO compute wait_map
        add_action(Weaver::Execution::Actions::WaitAction.new({}, :delay => time, :delete => @plan.last.is_a?(Weaver::Execution::Actions::DeleteNodesAction) ))
        # STDOUT.puts "updated plan = #{@plan.inspect}"
      end # def wait
      
      # Execute a change plan by executing each action in the plan
      # Actions are executed unless a wait action forces a pause
      # * *Args*    :
      # - +properties+
      # - +execution_parameters+
      #
      # * *Returns* : the updated topology resulting from the application of the plan
      def execute (properties, execution_parameters)
        # preprocess_plan

        # create coordination transaction
#        Weaver::Coordination::System.instance.set_property(:zookeeper_hosts, Weaver::System.instance.get_property(:zookeeper_hosts))
#        Weaver::Coordination::System.instance.set_property(:zookeeper_port, (Weaver::System.instance.get_property(:zookeeper_port) || 2181))
        Weaver::Coordination::System.instance.set_property(:log_level, Weaver::System.instance.get_property(:log_level))
          
        coordinator = Weaver::Deployers::Execution::CoordinationUtil::get_coordinator
        @transaction = coordinator.create_transaction()
        # @transaction = Weaver::Coordination::CoordinatedTransactionManager.instance.create_transaction(coordinator)
        @@logger.debug "Created transaction #{@transaction.inspect}"
        # populate the coordinator with environment details, if known
        Weaver::Deployers::Execution::CoordinationUtil.write_defaults(@transaction, @root)
        
        # execute actions as individual threads; they are expected to coordinate among themselves
        @plan.each do |action|
          action.execute_action(properties, execution_parameters)
        end # @plan.each do
        
        # TODO complete coordination transaction
        # Need to wait for action to really finish
#        @transaction.complete()
        
        # TODO schedule deletion of the transaction record: wait for VMs to be configured
        # perhaps each action should be blocking until this is done
#        @transaction.delete()
        
        return @root
      end # def execute
      
      def to_s
        return @plan.to_s
      end
      
      protected
      
#      # Records data about the environment in the coordinator including:
#      #   - deployment id, status for environment
#      #   - hostname, ip_address, status(es) of nodes
#      def write_defaults ()
#        [:deployment, :config_status].each do |attr_id|
#          attr = @root.topologies.values.first.amd[attr_id]
#          @transaction.write_data(attr.fully_qualified_id, attr.value) unless attr.nil? or attr.value.empty?
#        end # [].each do
#
#        @root.flatten_children.select {|c| c.is_a?(Weaver::Core::Node) && !Weaver::Transforms::Unroll.template?(c) }.each do |node|
#          # TODO Do we need to do this if it is already there?
#          [:ip_address, :hostname, :config_status, :vm_status].each do |attr_id|
#            attr = node.amd[attr_id]
#            @transaction.write_data(attr.fully_qualified_id, attr.value) unless attr.nil? or attr.value.nil? or attr.value.empty?
#          end # [].each do
##          @transaction.write_data(node.ip_address.fully_qualified_id, node.ip_address) unless node.ip_address.nil?
##          @transaction.write_data(node.hostname.fully_qualified_id, node.hostname) unless node.hostname.nil? 
#          # TODO elastic ip addresses
#          # TODO storage ?
##          @trasaction.write_data(node.config_status.fully_qualified_id, node.config_status) unless node.amd[:config_status].nil?
##          @transaction.write(node.vm_status.fully_qualified_id, node.vm_status) unless node.amd[:vm_status].nil?
#        end # @operation_map.each do
#      end # def write_defaults
#      
      # Adds action to plan by appending it to the plan
      def add_action (action)
        action.associate_with_plan(self, get_action_id)
        @plan << action
      end # end add_action
      
      def get_action_id
        @action_index  = @action_index + 1
        return "action_#{@action_index}"
      end # def get_action_id
      
      # Generate a Hash Weaver::Core::Node --> List<Automation> mapping nodes in a root to a list of automations.
      # Here the list of automations are those to be executed by a RunAutomationAction
      def create_operations_map (root, automation_tags = [], node_filter = [], deployed = true)
        filter = node_filter
        if filter.nil? or filter.empty?
          filter = root.flatten_children.select {|c| c.is_a?(Weaver::Core::Node) }.map { |n| n.fully_qualified_id }
        end
        node_filter = filter.map { |id| root.instance_eval(id) }
  
        automation_map = {}
        automations_to_run(root, :deployed => deployed, :nodes => node_filter, :automation_tags => automation_tags).each do |a|
          node = node(a)
          automation_map[node] = [*automation_map[node]] << a
        end # def automations_to_run
        
        return automation_map
      end # create_operations_map
  
      # Identify all automations to run from a root entity
      # Filter the set of automations based on:
      #   - The automation is on a node that is (or is not) already deployed (:deployed)
      #   - The automation is on a particular node (:nodes)
      #   - The automation has a particular tag (:automation_tags)
      def automations_to_run (root, filters = {})
        # find all automations
        @@logger.debug "filter: deployed? --> #{filters[:deployed]}"
        @@logger.debug "filter: nodes --> #{filters[:nodes].map {|n| n.fully_qualified_id}.inspect}"
        @@logger.debug "filter: automation tags --> #{filters[:automation_tags].inspect}"
        
        # all automations in the model
        all_automations = Set.new(root.flatten_children.select { |c| c.is_a?(Weaver::Core::Automation)})
        # puts "all_automations: #{all_automations.size}"
        # all_automations.each {|a| puts "all automations: #{a.fully_qualified_id} (#{deployed?(node(a))}); tags = #{a.tags.inspect}" }
          
        # automations filtered by @deployed flag
        filtered_by_deployment = all_automations.select { |a| deployed?(node(a)) == filters[:deployed] }
        # puts "filtered_by_deployment: #{filtered_by_deployment.size}"
          
        # automations filtered by tags
        filtered_by_tag = filtered_by_deployment.select { |a| a.tagged_with?(filters[:automation_tags]) }
        # puts "filtered_by_tag: #{filtered_by_tag.size}"
        # filtered_by_tag.each {|a| puts "candidate automations: #{a.fully_qualified_id}" }
          
        # automations filtered by node_filter; note special handling to deal with nodes in groups
        nodes = {}
        filtered_by_tag.each { |a| nodes[a] = node(a) }
        node_equivalents = nodes.merge(nodes) { |a,n| Weaver::Transforms::Unroll.group(n).nil? ? n : Weaver::Transforms::Unroll.template(Weaver::Transforms::Unroll.group(n))}
        # node_equivalents.each { |a,n| puts "#{a.fully_qualified_id} in #{nodes[a].id} ~ #{node_equivalents[a].id}" }
        filtered_by_node_filter = filtered_by_deployment.select { |a| filters[:nodes].include?(node_equivalents[a]) && !Weaver::Transforms::Unroll::template?(node(a))}
        # puts "filtered_by_node_filter: #{filtered_by_node_filter.size}"
  
        # remove any automations on template nodes
        final_automations = filtered_by_node_filter.each { |a| ! Weaver::Transforms::Unroll.template?(nodes[a]) } 
        # puts "@automations: #{@automations.size}"
        # final_automations.each {|a| puts "final automations: #{a.fully_qualified_id}; tags = #{a.tags.inspect}" }
          
        # add automations on which the set depends
        previous_size = 0
        while final_automations.size > previous_size
          previous_size = final_automations.size
          (all_automations - final_automations).each do |candidate|
            final_automations.each do |a|
              # TODO implement
  #              if a.depends_on(candidate) then
  #                all_automations << candidate
  #              end # if 
            end # automations.each do
          end # (all_automations - automations).each do
        end # while automations.size > previous_size
        
        # TODO validate that the filters[:nodes] requirement isn't violated ?
        return final_automations
      end
      
      # Retrieve the Weaver::Core::Node containing an entity (Base)
# TODO make the is a general function of node?
      def node (entity)
        while !entity.nil? and !entity.is_a?(Weaver::Core::Node)
          entity = entity.scope
        end
        return entity
      end # def node
      
      # Determine if a node is deployed
      # Currently validates the node and checks for an ip address
# TODO make the is a general function of node?
      def deployed?(node)
        !node.nil? and !node.ip_address.nil? and ![:storage].include?(node.tags)
      end # def deployed?
      
      DEFAULT_TEMPLATE_ID = :default

      # Retrieve the specified change template
      # TODO read from project folders or configuration database instead
      def retrieve_template (change_template_id)
        id = change_template_id.nil? ? DEFAULT_TEMPLATE_ID : change_template_id
        template = File.expand_path(File.join('./templates', "#{id}.template"), File.dirname(__FILE__))
        @change_template = File.open(template).read()
        @@logger.debug "Change template '#{id}' retrieved as: #{@change_template}"
      end # def retrieve_template
      
      # Specialize change template for a particular deployment
      def specialize_template (config = {})
        assignments =  "assignments {\n"
        assignments << config.map { |key, value| %Q{  @#{key.to_s} = #{value.inspect}} }.join("\n")
        assignments << "\n}"
        @change_template = assignments + "\n\n" + @change_template
        # puts "Change template updated as: #{@change_template}"
      end # def specialize_template
      
      # TODO fix this
      def identify_nodes_to_delete(nodes)
        return @desired_state.flatten_children.select do |child| 
          child.is_a?(Weaver::Core::Node) and 
#          child.realized_by.empty? and 
          !child.tags.member?(:storage) and 
          !Weaver::Transforms::Unroll.template?(child)
        end.select { |n| nodes.include?(n.fully_qualified_id) || nodes.empty? }
      end # def identify_nodes_to_delete
      
    end # class ChangePlan

  end # module Execution
end # module Weaver
