#!/usr/bin/env ruby
################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

################################################################################
# Dependencies                                                                 #
################################################################################
require "rubygems"
require "logger"
require "fileutils"
require "open-uri" # helper to open files regardles of whether it is on the local file system or at a URI
require "uri"
require 'net/http'
require "json"

################################################################################
# Class Definition                                                             #
################################################################################

class ChefUnpacker
  ROOT_PACKAGES_DIR = "/var/iwd/chef-solo"
  
  CT_TEXT_PLAIN = "text/plain"
  
  LOG_FILE = STDOUT
  LOG_LEVEL = Logger::DEBUG
  
  @packages # an array of pacakges to acquire
  @package_dirs
  @log
  
  def initialize
    @log = Logger.new(LOG_FILE)
    @log.level = LOG_LEVEL

    @packages = []
  end
  
  # Helper to clean up the the ROOT_PACKAGES_DIR from previous runs and recreate
  # it.
  #
  def generate_dir_structure
    if File.directory?(ROOT_PACKAGES_DIR)
      FileUtils.rm_rf ROOT_PACKAGES_DIR
    end
    
    FileUtils.mkpath(ROOT_PACKAGES_DIR)
  end

   # Unpack cookbook archive from bundle
   #
   def unpack_cookbooks
      @log.debug "Unpacking cookbook archives"

      Dir["*.tgz"].each do |pkg|
         @log.debug "   Unpacking #{pkg}"
         @log.debug "tar -xzvf #{pkg} -C #{ROOT_PACKAGES_DIR}"
         `tar -xzvf #{pkg} -C #{ROOT_PACKAGES_DIR}`
      end
   end
  
  # Create a roles directory in ROOT_PACKAGES_DIR and create symlinks to 
  # all of the role files in each chef package.
  #
  def create_roles_directory
    Dir.mkdir("#{ROOT_PACKAGES_DIR}/roles")
    
    Dir["#{ROOT_PACKAGES_DIR}/**/roles/*.rb"].each do |role_file_path|
      File.symlink(role_file_path, "#{ROOT_PACKAGES_DIR}/roles/#{File.basename(role_file_path)}")
    end
  end
  
  # Create a nodes directory in ROOT_PACKAGES_DIR and create symlinks to 
  # all of the node files in each chef package.
  #
  def create_nodes_directory(bundle)
    Dir.mkdir("#{ROOT_PACKAGES_DIR}/nodes")
    
    Dir["#{ROOT_PACKAGES_DIR}/**/nodes/*.json"].each do |node_file_path|
      File.symlink(node_file_path, "#{ROOT_PACKAGES_DIR}/nodes/#{File.basename(node_file_path)}")
    end
  end

   def create_node_file(bundle, overrides)
      node_file = bundle + "_node.json"
      node_file_loc = File.join(ROOT_PACKAGES_DIR+"/nodes", node_file)
      #unless File.exists?(node_file_loc)
         File.open(node_file_loc, 'w') { |file|
            file.puts JSON.pretty_generate(overrides.merge!({"run_list" => "role[#{bundle}]"}))
         }
      #end
      return node_file_loc
   end

  # Create a chef config file if one does not already exist
  #
  def create_chef_config(bundle)
    chef_config_file = "solo.rb"
    config_file_loc = File.join(ROOT_PACKAGES_DIR, chef_config_file)
    
    # create a solo.rb file even if it already exists
    # unless File.exists?(config_file_loc)
      File.open(config_file_loc, 'w') { |config_file|  
        config_file.puts "file_cache_path \"#{ROOT_PACKAGES_DIR}\""
        config_file.puts "role_path \"#{ROOT_PACKAGES_DIR}/roles\""
        config_file.puts "cookbook_path \"#{ROOT_PACKAGES_DIR}/cookbooks\""
      }
    # end
    return config_file_loc
  end

  def ChefUnpacker.run
    unpacker = ChefUnpacker.new

   role = ""
   arg_map = {}
   
   env_to_json_map = Hash.new()
   
   weaver_text = IO.read("role.weaver")
   if(weaver_text)
   		weaver_text.each_line { |line|
   	 		property_loc = line.index(/property :/)
   	 		if(property_loc)
   	 			property_loc = property_loc + 10
   	 			space_loc = line.index(' ', property_loc)
   	 			if(space_loc)
   	 				key = line[property_loc...space_loc]
   	 				name_loc = line.index(/:name => /, space_loc)
   	 				if(name_loc)
   	 				    name_loc = name_loc + 10
   	 					space_loc = line.index('"', name_loc)
   	 					if(space_loc)
   	 						value = line[name_loc...space_loc]
   	 						env_to_json_map[key] = value
   	 					end
   	 				end
   	 			end
   	 		end
   	 	}
   end
   
   ENV.to_hash.each do |key, value|
      if "ROLE_FILE" == key
         role = value
      else
      	 key_to_json = env_to_json_map[key]
      	 if(key_to_json)
         	keykeys = key_to_json.split('.')
         	key_major = keykeys.shift
         	a_map = arg_map
         	until keykeys.length == 0
            	key_minor = keykeys.shift
            	a_map[key_major] = {} if a_map[key_major].nil?
            	if (keykeys.length == 0)
               		a_map[key_major].merge!({key_minor => value})
            	else
               		a_map[key_major].merge!({key_minor => {}}) if a_map[key_major][key_minor].nil?
               		a_map = a_map[key_major]
               		key_major = key_minor
            	end
         	end # end of until...
         end # end of if(key_to_json)       
       end # end of else
   end # end of ENV processing

   until ARGV.length == 0
      key = ARGV[0].sub(/^-+/, '') 
      ARGV.shift
      if /^ROLE_FILE$/ =~ key
         role << ARGV[0]
      else
         keykeys = key.split('.')
         value = ARGV[0]
         key_major = keykeys.shift
         a_map = arg_map
         until keykeys.length == 0
            key_minor = keykeys.shift
            a_map[key_major] = {} if a_map[key_major].nil?
            if (keykeys.length == 0)
               a_map[key_major].merge!({key_minor => value})
            else
               a_map[key_major].merge!({key_minor => {}}) if a_map[key_major][key_minor].nil?
               a_map = a_map[key_major]
               key_major = key_minor
            end
         end
      end
      ARGV.shift
   end
   
   puts "role => #{role}\n"

   ROOT_PACKAGES_DIR << "/" + role
   #FileUtils.mkpath(ROOT_PACKAGES_DIR)
   unpacker.generate_dir_structure
   FileUtils.mkpath(ROOT_PACKAGES_DIR + "/cookbooks")
   unpacker.create_roles_directory
   unpacker.create_nodes_directory role

   unpacker.unpack_cookbooks
   # create solo.rb file
   config_file = unpacker.create_chef_config role
   # create node file w/ overrides
   node_file = unpacker.create_node_file(role,arg_map)

   return config_file, node_file
  end
end

################################################################################
# Main Method                                                                  #
################################################################################
if __FILE__ == $PROGRAM_NAME
    config_file, node_file = ChefUnpacker.run
    puts "Running: chef-solo -c #{config_file} -j #{node_file}"
    chef_out = `chef-solo -c #{config_file} -j #{node_file}`
    puts chef_out
    exit(0)
end