require 'weaver/transforms/duplicate'

module Weaver
  module Transforms
    include Weaver::Core
    # Implements Weaver 'topology expansion' based on node multiplicity.
    #
    # Given a Weaver entity (typically a {Root}), creates another Weaver entity.  In the new entity,
    # every {Topology} containing a {Node} with multiplicity N (and N > 1) will have N additional Nodes
    # and a {Group} linking the original Node and the created siblings with tag +:multiplicity+.  Nodes
    # that are members of the group will have an attribute +template+ which will be +true+ if the Node
    # should not be deployed but only describes a cluster of nodes.  Nodes with +template+ +false+ will
    # have a +multiplicity_index+ attribute, a FixNum in the range 1..N.
    #
    # Unroll.template(Node) returns +true+ for templates
    #
    # Unroll.group(Node) returns a Group for Nodes that are part of a scaling group
    #
    # Author:: Michael Kalantar
    class Unroll
      include Weaver::Core

      # @private
      # We avoid expanding the name and description because they are non-functional and
      # we wish to preserve the original names.  We avoid unrolling :ip_address and :hostname
      # because they aren't 'true' attributes but rather aliases for :nic0's attributes.
      UNROLLED_ATTRIBUTE_IDS = Set.new [:name, :description, :ip_address, :hostname ]

      MULTIPLICITY_TAG = :multiplicity

      # Create an entity logically equivalent but containing nodes with multiplicity 1
      # * +entity+ a Base, typically a Topology or Root, to be unrolled
      def self.unroll(entity)
        raise WeaverError.new("Can't unroll nil") if entity.nil?
        unroller = Unroll.new(entity)
        return unroller.unrolled_entity
      end # def self.unroll

      # The actual unrolled value
      def unrolled_entity
        return @entity
      end # def unrolled_entity

      protected

      # * +orig_entity+ a Base, typically a Topology or Root, to be unrolled
      def initialize(orig_entity)
        out = []
        @entity = Duplicate.duplicate(orig_entity)
        # process nodes that have a multiplicity > 1 or are templates
        @entity.flatten_children.select { |n| n.kind_of?(Weaver::Core::Node) and (n.multiplicity.kind_of?(Fixnum) && n.multiplicity > 1 or Unroll.template?(n))}.each do |node|
          group = sync_node(node)
        end # @entity.flatten_children
      end # def initialize

      # if node is already a template, find the existing group and modify the members if needed
      # the following assumptions are used:
      # (a) the value on the template is the correct value; the number of group members should be adjusted to comply
      # (b) when members need to be deleted from the group any nodes can be deleted
      def sync_node(node)
        # if node is a member of a group already then we shouldn't be here
        # get or create a group entity in the model

        group = Unroll.group(node, true)

        # add or delete nodes to make the size of the group match the desired size
        # remember that the group includes the template node
        while 0 != diff = node.multiplicity - group.members.size + 1 do
          if diff > 0 then
            add_node(group)
          else
            delete_node(group)
          end # if diff > 0
        end # while 0 != diff

        # if only 1 member left, should eliminate the group
        if 2 >= group.members.size
          # find the template (will delete it)
          template = Unroll.template(group)

          # remove references to group member(s)
          to_remove = group.members.map { |id, v| id }
          to_remove.each { |id| group.members.delete(id) }
          #delete template node from containg base
          template.scope.nodes.delete(template.id)
          # delete the group
          group.scope.groups.delete(group.id)
          # why is this here?
          # template.amd.delete(:multiplicity_index)
        end # if 2 >= group.members.size

      end # def sync_node

      # Identify the group associated with a node
      def self.group(node, create = false)
        contained_in = node.member_of.select { |m,g|  Unroll.multiplicity_group?(g) }.values
        return contained_in.first unless contained_in.empty?

        return nil unless create
        return Unroll.create_group(node)
      end # def group

      # Create a group using a node as a template
      def self.create_group(node)
        # create a multiplicity group
        node_scope = node.scope
        gid = "#{node.id}_mg".to_sym
        grp = Weaver::Core::Group.new(gid, node.scope)
        node.scope.groups[gid] = grp
        grp.name = node.name if node.name
        grp.description = node.description if node.description
        grp.amd[:member_index] = Weaver::Core::Attribute.new(:member_index, grp)
        grp.amd[:member_index].value = 0
        grp.tags  << MULTIPLICITY_TAG

        # add the node to the group, mark it as a template
        grp.members[node.id] = node
        node.amd[:template] = Weaver::Core::Attribute.new(:template, node)
        node.amd[:template].value = true

        return grp
      end # def create_group

      # Add a node to a group by duplicating the template and adding the result to the group.
      def add_node(group)
        # Assumes there is a template node in the group already
        template = Unroll.template(group)
        idx = group.member_index + 1
        copy_id = "#{template.id}_#{idx}".to_sym
        copy = Duplicate.duplicate(template)
        copy.id = copy_id
        copy.amd[:multiplicity].value = 1
        copy.amd[:template].value = false
        copy.amd[:multiplicity_index] = Weaver::Core::Attribute.new(:multiplicity_index, copy)
        copy.amd[:multiplicity_index].value = idx
        template.scope.nodes[copy_id] = copy
        group.members[copy_id] = copy
        group.amd[:member_index].value = idx
      end # def synchronize_group

      # Delete a node from a group. Nodes are deleted with this order of priority:
      #     (i) failed member nodes first
      #    (ii) undeployed member nodes second
      #   (iii) the oldest
      # If the node is penultimate node to be deleted, the group is deleted from the model.
      def delete_node(group,id = nil)
        # select node to delete
        node = select_node_to_delete(group, id)

        # remove the node from the group and delete it
        group.members.delete(node.id)
        node.scope.nodes.delete(node.id)
      end

      # Select a node to delete from a group
      def select_node_to_delete(group, id = nil)
        to_delete = nil

        # if a specific node was specified
        unless id.nil?
          to_delete = group.members.valuses.find { |v| v.id == id }
        end
        @@logger.debug "To delete: requested node '#{to_delete.id.inspect}'" unless to_delete.nil?
        return to_delete unless to_delete.nil?

        # find a failed node
        to_delete = group.members.values.find { |m| !Unroll.template?(m) and failed?(m) }
        @@logger.debug "To delete: failed node '#{to_delete.id.inspect}'" unless to_delete.nil?
        return to_delete unless to_delete.nil?

        # find an undeployed node
        to_delete = group.members.values.find { |m| !Unroll.template?(m) and !deployed?(m) }
        @@logger.debug "To delete: undeployed node '#{to_delete.id.inspect}'" unless to_delete.nil?
        return to_delete unless to_delete.nil?

        # find the oldest node
        to_delete = group.members.values.find { |m| !Unroll.template?(m) }
        @@logger.debug "To delete: oldest node '#{to_delete.id.inspect}'" unless to_delete.nil?
        return to_delete
      end # def delete_node

      # Retrieve the template for a group. There should be only 1, we return the first found.
      def self.template(group)
        return group.members.values.find { |member| Unroll.template?(member) }
      end # def template

      # Determine if a node has been unrolled.  When a node is unrolled, it becomes a template and is
      # included in a group. Both conditions are tested and must hold to return true.
      def self.unrolled?(node)
        if Unroll.template?(node)
          return !node.member_of.select { |m,g| Unroll.multiplicity_group?(g) }.empty?
        else return false
        end # if template?(node)
      end # def unrolled?

      # Determine if a node is a template node
      def self.template?(node)
        return false unless node.amd.include?(:template)
        return node.template
      end # def template?(node)

      # Determine if a node has been deployed. Test is whether or not an ip_address has been assigned
      def deployed?(node)
        return false if node.ip_address.nil?
        return true
      end # def deployed?(node)

      # Determine if a node failed its deployment
      # TODO: check vm_status as well as configuraiton status
      def failed?(node)
        return false unless node.amd.include?(:config_status)
        return node.config_status == "failed"
      end # def failed?(node)

      # Determine if a group is a multiplicity group (all groups of nodes should be)
      def self.multiplicity_group?(group)
        return false if group.nil?
        return group.tags.include?(MULTIPLICITY_TAG)
      end

    end # class Unroll

  end # module Transforms
end # module Weaver
