require 'weaver/core/logging'
require "devops_core/crypto_util"
require "base64"

module Weaver
  # \Weaver::Utils provides utility functions the \Weaver developers have found helpful.
  module Utils
    # This class decrypts {Weaver::Core::EncryptedValue}
    # 
    # Author:: Ed Snible
    module ValueDecryptor
      include Weaver::Core  # for logging
      
      # @private
      # a string: whitespace, "decrypt(", whitespace, quote, string, same-quote, whitespace, ")", whitespace
      DECRYPT_REGEXP = /\A\s*decrypt\(\s*(['"])(.*)\1\s*\)\s*\z/
      
      # * +encrypted_value+ a {Weaver::Core::EncryptedValue}
      # * +keyfile+ optional ezcrypto keyfile
      def self.decrypt(encrypted_value, keyfile = nil)
        
        keyfile = keyfile || ENV['KEY_FILE']
        @@logger.warn("No keyfile supplied; using default keyfile which is neither private nor provides good security") if keyfile.nil?
        
        unless keyfile.nil? || File.exist?(keyfile)
          raise WeaverError.new("KEY_FILE #{keyfile.inspect} does not exist")
        end
        
        encoded = Base64.decode64(encrypted_value.ciphertext)

        begin
          plaintext = DevopsCore::CryptoUtil.decrypt(encoded, keyfile)
        rescue OpenSSL::Cipher::CipherError => ce
          if keyfile.nil?
            raise WeaverError.new("Can't process #{encrypted_value.inspect} without defined $KEY_FILE")
          end
          raise WeaverError.new("Can't process #{encrypted_value.inspect} using $KEY_FILE #{keyfile.inspect}")
        end
        
        return plaintext
      end
      
      # * +propname+ to be included in warnings if there is no encryption
      # * +possibly_encrypted_string+ an expression like "decrypt('base64text') OR a plain string"
      # * +keyfile+ (optional) keyfile name
      def self.decrypt_if_needed(propname, possibly_encrypted_string, keyfile = nil)
        return nil if possibly_encrypted_string.nil?
        
        match = DECRYPT_REGEXP.match(possibly_encrypted_string)
        if match.nil?
          @@logger.warn "#{propname} is stored as plaintext"
          return possibly_encrypted_string
        end 
        
        keyfile = keyfile || ENV['KEY_FILE']
        encoded = Base64.decode64(match[2])
        begin
          plaintext = DevopsCore::CryptoUtil.decrypt(encoded, keyfile)
        rescue OpenSSL::Cipher::CipherError => ce
          if keyfile.nil?
            raise WeaverError.new("Can't process #{match[2].inspect} without defined $KEY_FILE")
          end
          raise WeaverError.new("Can't process #{match[2].inspect} using $KEY_FILE #{keyfile.inspect}")
        end
        return plaintext
      end
      
      # * +expr_string+ an expression like "decrypt('base64text')"
      def self.base64_from_expression(expr_string)
        match = DECRYPT_REGEXP.match(expr_string)
        return match[2]
      end
    end
  end
end
