require 'singleton'
require 'weaver/coordination/logging'
require 'weaver/coordination/system'
require 'weaver/coordination/coordinated_transaction'
require 'weaver/coordination/coordination_error'
require 'weaver/coordination/zk_shared_memory'

module Weaver
  module Coordination
    
    TRANSACTION_STATUS_NM = 'tx_status'
    module TRANSACTION_STATUS
      CREATED = 'created'
      STARTED = 'started'
      COMPLETED = 'completed'
      ABORTED = 'aborted'
    end
    
    class Coordinator
      include Weaver::Coordination::Logging
      
      SIMULATED = "__simulated_coordination_server__"
      WEAVER_TX_BASE = "weaver_coordinated_transactions"
      
      attr_reader :id
      attr_reader :hosts

      def initialize(id, servers, options = {})
        @hosts = [*servers]
#        @servers = [SIMULATED] if @servers.empty?
#        @port = port
#        @id = @servers.map { |server| "#{server}:#{@port}" }.join(',')
        @@logger.debug "options are: #{options.inspect}"
        @id = id
        @mock = false
        [:mock, :mock_coordination].each {|k| @mock = true if options[k] or "true" == options[k] }
        @simulate = false
        @simulate = true if options[:simulate]
        @zksm = nil
      end # def initialize
      
      # Create a new coordination transaction
      def create_transaction(options = {})
        # create id for transaction
        @@logger.debug "create_transaction called with options #{options.inspect}"
        tx_id = create_transaction_id(options)
        # create new client connection to zk for the transaction
        tx = Weaver::Coordination::CoordinatedTransaction.new(tx_id, self, options)
        tx.begin
        return tx
      end # def create_transaction
      
      # join an existing transaction
      def join_transaction(transaction_id, options = {})
        # verify that is is an existing transaction
        raise CoordinationError.new("Invalid transaction identifier: (nil)") if transaction_id.nil?
        
        # handle simulated coordination server case
        # always allow join to succeed
        if @simulate and !transaction?(transaction_id)
          expect(transaction_path(transaction_id, Weaver::Coordination::TRANSACTION_STATUS_NM), 
            Weaver::Coordination::TRANSACTION_STATUS::STARTED)
        end
        
        raise CoordinationError.new("Transaction not created: '#{transaction_id}'") unless transaction?(transaction_id)
        status = transaction_status(transaction_id)
        raise CoordinationError.new("Transaction completed: '#{transaction_id}'") if Weaver::Coordination::TRANSACTION_STATUS::COMPLETED == status
        raise CoordinationError.new("Transaction aborted: '#{transaction_id}'") if Weaver::Coordination::TRANSACTION_STATUS::ABORTED == status
        tx = Weaver::Coordination::CoordinatedTransaction.new(transaction_id, self, options)
        return tx
      end # def join_transaction
      
      def mock!
        @mock = true
        # if switching to mocking, then essentially redefining the coordinator
        # make sure we clear our reference to shared memory
        @zksm = nil
      end # def mock!
      
      def unmock!
        @mock = false
        # if switching from mocking, then essentially redefining the coordinator
        # make sure we clear our reference to shared memory
        @zksm = nil
      end # def unmock!
      
      def mock?
        return @mock
      end # def mock?
      
      def expect(path, value)
        ZKMockService.instance.expect(@id, path, value) unless !mock?
      end # def expect
      
      def transaction_path(id, path = nil)
        path.sub!(/^\//, '') unless path.nil?
        tx_path = path.nil? ? "/#{WEAVER_TX_BASE}/#{id}" : "/#{WEAVER_TX_BASE}/#{id}/#{path}"
        return tx_path
      end # def transaction_path
      
      def can_coordinate?
        return !@simulate
      end # def can_coordinate?

      protected
      
      # Generate a coordination transaction identifier
      # Option :transaction_prefix can be used to prefix the id so that it is more easily identified
      # For legacy, supports use of global system property as well
      def create_transaction_id (options = {})
        prefix = options[:transaction_prefix] || Weaver::Coordination::System.instance.get_property(:transaction_prefix) || nil
        suffix = SecureRandom.uuid
        return prefix.nil? ? suffix : "#{prefix}##{suffix}"
      end # def create_transaction_id
      
      # determine if a transaction w/ a particular id exists
      def transaction? (transaction_id)
        return false if transaction_id.nil?
        connect
        return @zksm.zk.exists?(transaction_path(transaction_id))
      end # transaction?
      
      # determine the status of a transaction
      def transaction_status (transaction_id)
        connect
        return @zksm.read(transaction_path(transaction_id, Weaver::Coordination::TRANSACTION_STATUS_NM))
      end # def transaction_status
      
      def connect
        @zksm = ZKSharedMemory.new(@id, {:mock => self.mock?}) if @zksm.nil?
        @zksm.zk.reopen unless @zksm.zk.connected?
      end # def connect
      
    end # class Coordinator
    
    class CoordinatedTransactionManager
      include Singleton
      include Weaver::Coordination::Logging
      
      # Establish connection to coordinator service (Apache Zookeeper)
      # *Args*
      #   - +coordinators+:: list of coordination (Zookeeper) servers (IP addresses)
      #   - +coordination_port+:: port used by client to connect to coordination servers
      # *Raises*
      #    Exception if cannot connect to the Apache Zookeeper
      def initialize ()
        setup_logging
        @coordinators = {}
        @default_coordinator = original_default()  
        @@logger.debug "default_coordinator is #{@default_coordinator.inspect}"
        add(@default_coordinator)
      end # def initialize
      
      def add_coordinator (servers = [], port = 2181, options = {})
        @@logger.debug "add_coordinator called with #{servers}:#{port}(#{options.inspect})"
        id = coordinator_id(servers, port);
        @@logger.debug "id is #{id}"
        @@logger.debug @coordinators.inspect
        return @coordinators[id] unless @coordinators[id].nil?
        return add(create_coordinator(id, [*servers], options))
      end # add_coordinator 
      
      def get_coordinator (servers = [], port = 2181)
        id = coordinator_id(servers, port)
        return id.nil? ? nil : @coordinators[id]
      end # get_coordinator
      
      def default_coordinator
        return @default_coordinator
      end # def default_coordinator
      
      def set_default_coordinator (coordinator)
        return if coordinator.nil?
        return unless coordinator.is_a?(Weaver::Coordination::Coordinator)
        @default_coordinator = coordinator
      end # set_default_coordinator
      
      def create_transaction(coordinator = nil, options = {})
        coordinator = @default_coordinator if coordinator.nil?
        return coordinator.create_transaction(options)
      end # def create_transaction
      
      def join_transaction(transaction_id, coordinator = nil)
        coordinator = @default_coordinator if coordinator.nil?
        return coordinator.join_transaction(transaction_id)
      end # def join_transaction
      
      # Determine whether or not the coordinator is a *real* coordinator exists; 
      # That is, it is not a mocked coordinator
      def can_coordinate?(coordinator = nil)
        coordinator = @default_coordinator if coordinator.nil?
        return coordinator.can_coordinate?
      end
      
      def mock!(coordinator = nil)
        coordinator = @default_coordinator if coordinator.nil?
        coordinator.mock!
      end # def mock!
      
      def unmock!(coordinator = nil)
        coordinator = @default_coordinator if coordinator.nil?
        coordinator.unmock!
      end # def mock!

      def expect(path, value, coordinator = nil)
        coordinator = @default_coordinator if coordinator.nil?
        return unless coordinator.mock? # TODO throw exception?
        coordinator.expect(path, value)
      end
            
      protected 
      
      def setup_logging
        levels = {"debug" => Logger::DEBUG, "info" => Logger::INFO, "warn" => Logger::WARN, "error" => Logger::ERROR, "fatal" => Logger::FATAL}
        log_level = "fatal"
        log_level = Weaver::Coordination::System.instance.get_property(:log_level) unless Weaver::Coordination::System.instance.get_property(:log_level).nil?
        level = levels[log_level.downcase]
        level = Logging::FATAL if level.nil?
        @@logger.level  = level
        set_level = {Logger::DEBUG => "debug", Logger::INFO => "info", Logger::WARN => "warn", Logger::ERROR => "error", Logger::FATAL => "fatal"}[@@logger.level]
        @@logger.info %Q{Coordinator logging level set to: #{set_level}}
        Weaver::Coordination::System.instance.set_property(:log_level, set_level)
      end # def setup_logging

      def original_default
        servers = [*Weaver::Coordination::System.instance.get_property(:zookeeper_hosts)]
        id = coordinator_id(servers, (Weaver::Coordination::System.instance.get_property(:zookeeper_port) || 2181))
        # create coordinator but make sure to mock if it is the simulated coordinator
        # this is the only place the simulated coordinator should be created
        options = Weaver::Coordination::Coordinator::SIMULATED == id ? {:mock => true, :simulate => true} : {}
        return create_coordinator(id, servers, options)
      end
      
      def coordinator_id (servers = [], port = 2181)
        id = [*servers].map { |server| "#{server}:#{port}" }.join(',')
        id = Weaver::Coordination::Coordinator::SIMULATED if id.nil? or '' == id
        return id 
      end # coordinator_id
      
      def create_coordinator(id, servers, options = {})
        @@logger.debug "create_coordinator, options = #{options.inspect}"
        return nil if id.nil? 
        return @coordinators[id] unless @coordinators[id].nil?
        coordinator = Coordinator.new(id, servers, options)
        return coordinator
      end
        
      def add(coordinator = nil)
        return nil if coordinator.nil?
        @coordinators[coordinator.id] = coordinator
        return coordinator
      end
      

  end # class EnvironmentAttributeStore
    
  end # module Coordination
end # module Weaver