require 'zk'
require 'weaver/coordination/zk_mock'
require 'weaver/coordination/logging'
require 'weaver/coordination/system'

module Weaver
  module Coordination
    
    # Client interface for distributed shared memory wrapping Zookeeper.
    class ZKSharedMemory
      include Weaver::Coordination::Logging
      
      INTERRUPT_BASE = "/interrupts"
      DEFAULT_SLEEP_TIME = 1;
      
      attr_accessor :zk
      
      # Create a client to the distributed shared memory
      # Params:
      # +zk_servers+:: list of Zookeeper servers
      # +mock+:: flag indicating whether to mock the zookeeper service
      def initialize(zk_servers, options = {})
        @queue = Queue.new
        @servers = zk_servers
        @mock = options[:mock] || options[:mock_coordinator] || "true" == (Weaver::Coordination::System.instance.get_property :mock_coordinator)
        @@logger.debug "mock = #{@mock} for zk servers: #{@servers}"
        @zk = @mock ? ZKMockService.instance.create(@servers) : ZK.new(@servers, :thread => :per_callback) #, :retry_duration => 60)
        # @zk.on_connected { |event| puts "connect handler called" }
      end # def initialize
            
      # Write string to distributed shared memory.
      # Params:
      # +path+:: location that should be written
      # +value+:: value to write
      def write (path, value)
        with_retries do
        # validate input: not nil
        raise ArgumentError, "Path must be non-null" if path.nil?
        # validate input: is a path; ie, starts with "/"
        raise ArgumentError, "Path must be begin with" unless path =~ /^\//

        # Until we figure out how the states really work, we avoid checking them
        # confirm_connection
        
        @@logger.debug "zksm write: #{path} = #{value}"
        if (!@zk.exists?(path))
          # create the path then write to it
          @zk.mkdir_p(dirname(path))
          @zk.create(path, value.to_s)
        else
          # location already exists; just write the value
          @zk.set(path,value.to_s)
        end # if (!@zk.exists?(path)) 
        end # with_retries
      end # def write
      
      # Utility method to interrupt a read
      def interrupt_read (id)
        with_retries do
          @zk.mkdir_p(interrupt_path(id))
        end # with_retries
      end # interrupt_read

      def close!
        @zk.close!
      end # close!
      
      def close
        @zk.close
      end

      def rm_rf (path)
        with_retries do
          @zk.rm_rf(path)
        end # with_retries
      end # def rm_rf

      # read method closest to Zookeeper
      def read (path, opts={})
        with_retries do
        # validate input: not nil
        raise ArgumentError, "Path must be non-null" if path.nil?
        # validate input: is a path; ie, starts with "/"
        raise ArgumentError, "Path must be begin with" unless path =~ /^\//

        # Until we figure out how the states really work, we avoid checking them
        # confirm_connection

        @@logger.debug "zksm read from: #{path}"
        
        # read options
        wait = opts[:wait].nil? ? true : opts[:wait]
        timeout = opts[:timeout].nil? ? 0 : opts[:timeout]
        wait_for_change = opts[:wait_for_change].nil? ? false : true

        # Create an event handler that contains a block of code that will be executed on an event triggered by a watch 
        eh = @zk.register(path) do |event|
          if event.node_changed? or event.node_created?
            data = event.zk.get(event.path)
            if !block_given? or yield data.first
              @queue.push(data)
            else
              # reset watch; check result in case set in the meantime
              data = event.zk.get(event.path, :watch => true)
              if !block_given? or yield data.first
                @queue.push(data)
              end # if yield data
            end # if yield
          end # if event.node_changed?...
        end # eh = @zk.register

#        seh = @zk.register_state_handler(:connected) do |event|
#        end # seh
        
   begin
        # If the path already exists, read the value. Otherwise, wait.
        # Then decide whether the value is acceptable or not; if not wait.
        if @zk.exists?(path, :watch => true)
          data = @zk.get(path)
          unless wait
            eh.unsubscribe
            return data.first
          end
          if wait_for_change or (block_given? and !(yield data.first))
            data = interruptable_wait path, create_id, timeout
          end
        else
          # watch for changes
          unless wait
            eh.unsubscribe
            return nil
          end
          data = interruptable_wait path, create_id, timeout
        end

        # Once we have the data, unsubscribe the event handler
        eh.unsubscribe

        @@logger.debug "zksm read: returning '#{data.first}'"
   rescue Exception => e
     eh.unsubscribe
     raise e
   end
        return data.first
        end # with_retries
      end # def read

      protected
     
      # We don't really understand the states so we've commented out the calls to this method
      RECONNECT_WAIT_TIME = 180

      def confirm_connection

        # return immediately if connected        
        return if @zk.connected?
        
        # wait for RECONNECT_WAIT_TIME seconds to reconnect
        @zk.reopen(RECONNECT_WAIT_TIME)
        
        # Still not connected?
        unless @zk.connected?
          raise ZK::Exceptions::ConnectionLoss.new('Unable to re-establish connection to zk (waited #{RECONNECT_WAIT_TIME} s)')
        end
      end # def confirm_connection

      # Creates a random string
      protected

      def create_id
        (0...8).map{65.+(rand(26)).chr}.join
      end

      # Waits for data to appear on a queue.
      # * Can be configured to timeout
      # * Can be interrupted by call to interrupt_read
      # Params:
      # +id+:: identifier associated with the read
      # +timeout+:: time alloted to the read in seconds if greater than 0.
      protected

      def interruptable_wait (path, id, timeout=0)
        @@logger.info "interruptable_wait on #{path} has is '#{id}'"
        begin
          @zk.rm_rf(interrupt_path(id))
        rescue ZK::Exceptions::NoNode
        end
        iterations = 0
        data = nil
        watcher = @mock ? ZKMock::MockWatcher.new() : Zookeeper::Callbacks::WatcherCallback.new {}
        interrupted = @zk.exists?(interrupt_path(id), :watcher => watcher)
        until !data.nil? or interrupted or
          (timeout > 0 and iterations*DEFAULT_SLEEP_TIME > timeout) do
          sleep DEFAULT_SLEEP_TIME
          begin
            # check for data to appear on queue
            data = @queue.pop(true)
          rescue ThreadError
          end
          interrupted = watcher.completed?
          iterations = iterations + 1
        end # until !data.nil?

        if interrupted
          begin
            @zk.rm_rf(interrupt_path(id))
          rescue ZK::Exceptions::NoNode
          end
          raise "Read call interrupted"
        end # if interrupted
        if data.nil? # timed out
          raise "Read call timed out"
        end
        data
      end # def interruptable_wait

      def uninterruptable_wait (path, id, timeout=0)
        @@logger.info "uninterruptable_wait on #{path}"# has is '#{id}'"
        data = @queue.pop()#true)
        if data.nil? # timed out
          raise "Read call timed out"
        end
        return data
      end # def uninterruptable_wait

      # Calculate the directory path to a location
      # Params:
      # +path+::
      def dirname (path)
        path = "/" if path.nil?
        dirname = path.sub(/\/[^\/]*$/, "")
        dirname = "/" if 0 == dirname.length
        dirname
      end # def dirname

      protected
      # Calculate the path to an interrupt indicator
      # Params:
      # +id+:: read identifier
      def interrupt_path id
        "#{INTERRUPT_BASE}/#{id}"
      end # def interrupt path
      
      MAX_CONNECTION_ATTEMPTS = 100
      RETRY_INTERVAL = 5
          
      def with_retries (&block)
        tries = 0
        success = false
        while !success # and tries < MAX_CONNECTION_ATTEMPTS
          begin
            result = yield
            success = true
            @@logger.debug "with_retries returning #{result}"
            return result
          rescue Zookeeper::Exceptions::NotConnected, ZK::Exceptions::ConnectionLoss => e
            @@logger.warn "Failed to communicate with coordinator: #{e}"
            sleep RETRY_INTERVAL
            tries += 1
            raise e if tries >= MAX_CONNECTION_ATTEMPTS
          end # begin.. rescue
        end # while !success
      end # def with_retries  

    end # class zk_shared_memory

  end # module Coordination
end # module Weaver